import java.applet.Applet;
import java.applet.AppletContext;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Iterator;
import java.util.Random;

/**
 * This code functions as a Java Application, a SatLab test harness jar, and a Java Applet.
 * It emulates BitTorrent and TCP traffic. If the transfer succeeds, it outputs the achieved throughput.
 * It communicates with a server program that acts as a listener and peer.
 * 
 * 
 * How to sign a java applet:
 *  - Generate a self-signed certificate:
 *       keytool -genkey -alias broadband.mpi-sws.mpg.de
 *  - jar the applet
 *       jar -cvf applet.jar BitTorrentClient*.class
 *  - sign the jar
 *        jarsigner applet.jar broadband.mpi-sws.mpg.de
 *
 *
 *  Known Problems: Detecting if a connection was resetted is not trivial in Java. Java just
 *                  throws a IOException if read() or write() fails. We have to parse the
 *                  Exception text. However, some versions of Java will localize this message.
 */

public class BitTorrentClient extends Applet {

	public static String sysinfo = "";
	public String errorpage = "bb/error.php";
	public String testpage = "bb/bt.php";

	protected BitTorrentWorker btw = null;
	protected InetSocketAddress PlServer = null;
	protected InetSocketAddress PlServer2 = null;
	protected static byte[] myID;
	protected String userid = "unknown";
	protected boolean up = false, down = false, tcp = false;
	protected int repeat = 1;
	protected int duration = -1;
	protected StringBuffer buffer;

	public boolean finished = false;

	public AppletContext appletContext = null;

	protected static int done = 0;
	protected int totalLength = 100;

	/** Thread that is used to calculate and display a progress bar to
	 *  show the finish time of the measurement
	 */

	private static class Countdown extends Thread {
		private BitTorrentClient parent = null;
		private long startTime;
		private long endTime;
		private int totalLength;
		private int bound;
		private int timeLeft;
		private boolean stalled = false;
		public boolean isTerminated = false;

		public Countdown(BitTorrentClient p, int totalLength) {
			super();
			this.startTime = System.currentTimeMillis();
			this.parent = p;
			this.totalLength = totalLength;
			this.bound = 0;
			this.timeLeft = totalLength;
		}

		public void setBound(int newBound) {
			this.bound = newBound;
			endTime = (timeLeft * 1000) + System.currentTimeMillis();
		}

		public void addToBound(int addBound) {
			this.bound += addBound;
			endTime = (timeLeft * 1000) + System.currentTimeMillis();
		}

		public void run() {
			endTime = startTime + (totalLength * 1000);

			while (!isTerminated && (endTime > System.currentTimeMillis())) {

				timeLeft = (int) ((endTime - System.currentTimeMillis()) / 1000.0);
				if (timeLeft < (totalLength - bound)) {
					timeLeft = totalLength - bound;
					stalled = !stalled;
				} else
					stalled = false;

				done = (int) (totalLength - timeLeft);
				if (!parent.finished) {

					if (!stalled)
						parent.printItem("Time to finish: "
								+ String.valueOf(timeLeft) + " seconds");
					else
						parent.printItem("Waiting for transfer to finish.");
				}
				try {
					sleep(1000);
				} catch (InterruptedException e) {
				}
			}

			done = (int) totalLength;
			if (!parent.finished)
				parent.printItem("Processing data...");
		}
	}

	private static class BitTorrentWorker extends Thread {

		protected BitTorrentClient btc = null;
		protected Countdown cd = null;
		protected InetSocketAddress server = null;
		protected InetSocketAddress server2 = null;
		protected InetSocketAddress local = null;
		protected boolean isTerminated = false;
		protected boolean resetted = false;

		protected final int commandPort = 19980;
		protected final int pieceSize = 262144;
		protected final int progressReportAfterBytes = 100000;

		protected long duration = 20000; // in milliseconds
		protected long setupDuration = 10000; // in milliseconds
		protected long pause = 4000; // in milliseconds - time between two experiments
		private int totalDuration = 0; // in seconds

		protected boolean transferSuccessful = true;
		protected long start, end;

		// Control variables
		protected int numberOfRepeats = 1;
		protected boolean setupUpstream = false;
		protected boolean setupDownstream = false;
		protected boolean setupUpstream2 = false;
		protected boolean setupDownstream2 = false;
		protected int tcpUpstream = 0;
		protected int tcpDownstream = 0;
		protected int btUpstream = 0;
		protected int btDownstream = 0;

		private PrintStream log = null;
		private String backLog = "";
		private Random rnd;

		protected final int readBufSizeBytes = pieceSize + 16 * 1024;
		protected final int writeBufSizeBytes = pieceSize + 16 * 1024;
		protected ByteBuffer rbuf;
		protected ByteBuffer wbuf;
		protected byte[] readBuf;
		protected byte[] writeBuf;
		protected int readLen;
		protected int writePtr;
		protected int writeLen;

		public void setFileOutput(String filename) {

			if (filename == null)
				return;

			try {
				log = new PrintStream(new FileOutputStream(filename), true);

				System.setOut(log);
				System.setErr(log);

			} catch (FileNotFoundException e) {
				System.err.println("Cannot open file " + filename);
				e.printStackTrace();
				log = null;
				return;
			}
		}

		/* Methods for encoding an integer in a byte array (Java was never good at this!) */

		protected int decodeInt(byte[] b, int offset) {
			return ((b[offset] & 0xFF) << 24) | ((b[offset + 1] & 0xFF) << 16)
			| ((b[offset + 2] & 0xFF) << 8) | (b[offset + 3] & 0xFF);
		}

		protected void encodeInt(int value, byte[] b, int offset) {
			b[offset] = (byte) ((value >> 24) & 0xFF);
			b[offset + 1] = (byte) ((value >> 16) & 0xFF);
			b[offset + 2] = (byte) ((value >> 8) & 0xFF);
			b[offset + 3] = (byte) (value & 0xFF);
		}

		public BitTorrentWorker(String server, int port) {
			this.server = new InetSocketAddress(server, port);

			rnd = new Random();
			myID = new byte[20];
			rnd.nextBytes(myID);
			rbuf = ByteBuffer.allocate(readBufSizeBytes);
			wbuf = ByteBuffer.allocate(writeBufSizeBytes);
			readBuf = rbuf.array();
			writeBuf = wbuf.array();
			readLen = 0;
			writePtr = 0;
			writeLen = 0;
		}

		public BitTorrentWorker(InetSocketAddress server) {
			this.server = server;

			rnd = new Random();
			myID = new byte[20];
			rnd.nextBytes(myID);
			rbuf = ByteBuffer.allocate(readBufSizeBytes);
			wbuf = ByteBuffer.allocate(writeBufSizeBytes);
			readBuf = rbuf.array();
			writeBuf = wbuf.array();
			readLen = 0;
			writePtr = 0;
			writeLen = 0;
		}

		public BitTorrentWorker(InetSocketAddress server, BitTorrentClient btc) {
			this.server = server;
			this.btc = btc;

			rnd = new Random();
			myID = new byte[20];
			rnd.nextBytes(myID);
			rbuf = ByteBuffer.allocate(readBufSizeBytes);
			wbuf = ByteBuffer.allocate(writeBufSizeBytes);
			readBuf = rbuf.array();
			writeBuf = wbuf.array();
			readLen = 0;
			writePtr = 0;
			writeLen = 0;
		}

		/**
		 * Given command line parameters, calculate how long the whole measurement run will take
		 * and set up the internal variables to control the measurement
		 */
		public void confExp(boolean up, boolean down, boolean tcp, int repeat, InetSocketAddress plServer2, int duration) {

			if (duration > 0) {
				this.duration = duration * 1000;
				this.setupDuration = Math.min(this.duration / 2, 10000);
				this.setupDuration = Math.max(this.setupDuration, 4000);
			}

			if (up) {
				this.btUpstream += repeat;
				if (tcp) {
					this.setupUpstream = true;
					this.tcpUpstream += repeat;
					totalDuration += 10;
				}
			}
			if (down) {
				this.btDownstream += repeat;
				if (tcp) {
					this.setupDownstream = true;
					this.tcpDownstream += repeat;
					totalDuration += 10;
				}
			}
			if (plServer2 != null) {
				this.tcpUpstream *= 2;
				this.tcpDownstream *= 2;
				this.btUpstream *= 2;
				this.btDownstream *= 2;
				this.server2 = plServer2;
				if (down)
					this.setupDownstream2 = true;
				if (up)
					this.setupUpstream2 = true;
			}

			if (repeat > 0)
				this.numberOfRepeats = repeat;

			totalDuration += (int) ((this.pause + this.duration) / 1000) * (this.btUpstream + this.btDownstream + this.tcpUpstream + this.tcpDownstream);

			if (btc != null) {
				cd = new Countdown(btc, totalDuration);
				btc.totalLength = totalDuration;
				cd.start();
			}
		}

		public void confExp(boolean up, boolean down, boolean tcp, int repeat, InetSocketAddress plServer2) {
			this.confExp(up, down, tcp, repeat, plServer2, -1);
		}

		public void confExp(boolean up, boolean down, boolean tcp, int port2, int repeat, int duration) {
			if (port2 > 0)
				this.confExp(up, down, tcp, repeat, new InetSocketAddress(server.getAddress(), port2), duration);
			else
				this.confExp(up, down, tcp, repeat, null, duration);
		}

		/**
		 * Creates a non-blocking SocketChannel with ReuseAddress set
		 * @param peer The server to connect to
		 * @param timeout The socket timeout in millis
		 * @return
		 */
		public SocketChannel setupSocket(InetSocketAddress peer, int timeout) {
			SocketChannel sChannel = null;
			Socket socket = null;

			/*
			if(local != null) {
				try{
					local = new InetSocketAddress(local.getPort()+1);
				} catch (IllegalArgumentException e) {
					local = null;
				}
			}
			 */

			try {
				sChannel = SocketChannel.open();
				sChannel.configureBlocking(true);
				socket = sChannel.socket();
				socket.setReuseAddress(true);
			} catch (IOException e2) {
				System.err.println("Failed to create new socket: " + e2);
				e2.printStackTrace();
				//handleFatalError("Cannot create socket", "id=" + btc.userid + "&msg=Cannot%20create%20socket", 8);
				return null;
			}

			try {
				socket.setSoTimeout(timeout);
				if (local != null)
					sChannel.socket().bind(local);
				sChannel.connect(peer);
				sChannel.configureBlocking(false);
			} catch (SocketException e2) {
				if (btc != null) {
					btc.printItem("Error: Connection to server "
							+ peer.toString() + " timed out.");
				} else {
					System.err.println("Connection timed out.");
				}
				try {
					sChannel.close();
				} catch (IOException e) {
				}
				return null;
			} catch (IOException e) {
				if (btc != null) {
					btc.printItem("Cannot connect to " + peer.toString());
				} else {
					System.err.println("Cannot connect to " + peer.toString()
							+ ". " + e);
				}
				return null;
			}

			/*
			if(local == null)
				local = new InetSocketAddress(socket.getLocalPort());
			 */
			return sChannel;
		}

		public void putString(byte[] buf, int offset, String str) {
			try {
				byte[] bytes = str.getBytes("US-ASCII");
				for (int i = 0; i < bytes.length; i++)
					buf[offset + i] = bytes[i];
			} catch (UnsupportedEncodingException uee) {
			}
		}

		public boolean checkString(byte[] buf, int offset, String str) {
			try {
				byte[] bytes = str.getBytes("US-ASCII");
				for (int i = 0; i < bytes.length; i++) {
					if (buf[offset + i] != bytes[i])
						return false;
				}
			} catch (UnsupportedEncodingException uee) {
			}

			return true;
		}

		public void chompRead(int numBytes) {
			for (int i = numBytes; i < readLen; i++)
				readBuf[i - numBytes] = readBuf[i];

			readLen -= numBytes;
		}

		public byte getTrulyRandomByte() {
			byte[] b = new byte[1];
			rnd.nextBytes(b);
			return b[0];
		}

		public int getTrulyRandomInt() {
			return rnd.nextInt();
		}

		protected int getInt(byte[] b, int offset) {
			return ((b[offset] & 0xFF) << 24) | ((b[offset + 1] & 0xFF) << 16)
			| ((b[offset + 2] & 0xFF) << 8) | (b[offset + 3] & 0xFF);
		}

		protected void putInt(byte[] b, int offset, int value) {
			b[offset] = (byte) ((value >> 24) & 0xFF);
			b[offset + 1] = (byte) ((value >> 16) & 0xFF);
			b[offset + 2] = (byte) ((value >> 8) & 0xFF);
			b[offset + 3] = (byte) (value & 0xFF);
		}

		public void fillWithTrulyRandomBytes(byte[] buf, int offset, int len) {
			byte[] b = new byte[len];
			rnd.nextBytes(b);
			for (int i = 0; i < len; i++)
				buf[offset + i] = b[i];
		}

		public void fillWithPseudoRandomBytes(byte[] buf, int offset, int len) {
			final int randomnessSize = 100;
			final int[] randomness = new int[] { 0x67, 0xC6, 0x69, 0x73, 0x51,
					0xFF, 0x4A, 0xEC, 0x29, 0xCD, 0xBA, 0xAB, 0xF2, 0xFB, 0xE3,
					0x46, 0x7C, 0xC2, 0x54, 0xF8, 0x1B, 0xE8, 0xE7, 0x8D, 0x76,
					0x5A, 0x2E, 0x63, 0x33, 0x9F, 0xC9, 0x9A, 0x66, 0x32, 0x0D,
					0xB7, 0x31, 0x58, 0xA3, 0x5A, 0x25, 0x5D, 0x05, 0x17, 0x58,
					0xE9, 0x5E, 0xD4, 0xAB, 0xB2, 0xCD, 0xC6, 0x9B, 0xB4, 0x54,
					0x11, 0x0E, 0x82, 0x74, 0x41, 0x21, 0x3D, 0xDC, 0x87, 0x70,
					0xE9, 0x3E, 0xA1, 0x41, 0xE1, 0xFC, 0x67, 0x3E, 0x01, 0x7E,
					0x97, 0xEA, 0xDC, 0x6B, 0x96, 0x8F, 0x38, 0x5C, 0x2A, 0xEC,
					0xB0, 0x3B, 0xFB, 0x32, 0xAF, 0x3C, 0x54, 0xEC, 0x18, 0xDB,
					0x5C, 0x02, 0x1A, 0xFE, 0x43 };

			for (int i = 0; i < len; i++)
				buf[offset + i] = (byte) randomness[i % randomnessSize];
		}

		public void log2(String x) {
			System.out.println(x);
		}

		public void log2i(String x, int i) {
			System.out.println(x+" "+i);
		}

		public double connectionHandler(SocketChannel sChannel,	Selector selector, long timeout, boolean useRealProtocol, boolean isSeeder) {
			long endTime = System.currentTimeMillis() + timeout;
			int state = isSeeder ? 0x00 : 0x10;
			int statLastIndex = 0;
			int statBytesTransferred = 0;
			int statPiecesTransferred = 0;

			transferSuccessful = false;
			resetted = false;
			readLen = 0;
			writePtr = 0;
			writeLen = 0;

			boolean protocolError = false;
			boolean connectionClosed = false;

			while (!protocolError && !connectionClosed) {

				/**********************************************************************************************************
				 * BitTorrent state machine. This is also valid Java code, and can be copied 1:1 to                       *
				 * BitTorrentClient.java                                                                                  */

				// The states of the BitTorrent protocol, separated by Seeder (start at 0x00) and Leecher (start at 0x10)
				final int STATE_SEEDER_WAIT_HANDSHAKE = 0x00;
				final int STATE_SEEDER_SEND_HANDSHAKE = 0x01;
				final int STATE_SEEDER_WAIT_BITFIELD = 0x02;
				final int STATE_SEEDER_SEND_BITFIELD = 0x03;
				final int STATE_SEEDER_WAIT_INTERESTED = 0x04;
				final int STATE_SEEDER_SEND_UNCHOKE = 0x05;
				final int STATE_SEEDER_WAIT_REQUEST_OR_CANCEL = 0x06;
				final int STATE_SEEDER_SEND_PIECE = 0x07;
				final int STATE_LEECHER_INIT_HANDSHAKE = 0x10;
				final int STATE_LEECHER_SEND_HANDSHAKE = 0x11;
				final int STATE_LEECHER_WAIT_HANDSHAKE = 0x12;
				final int STATE_LEECHER_SEND_BITFIELD = 0x13;
				final int STATE_LEECHER_WAIT_BITFIELD = 0x14;
				final int STATE_LEECHER_SEND_INTERESTED = 0x15;
				final int STATE_LEECHER_WAIT_UNCHOKE = 0x16;
				final int STATE_LEECHER_SEND_REQUEST = 0x17;
				final int STATE_LEECHER_WAIT_PIECE = 0x18;

				switch (state) {
				case STATE_LEECHER_INIT_HANDSHAKE: {
					/* The leecher starts by sending a handshake message. We make up random hash and ID values */

					if (useRealProtocol) {
						for (int i = 0; i < 68; i++)
							writeBuf[i] = 0;
						writeBuf[0] = 19;
						putString(writeBuf, 1, "BitTorrent protocol");
						for (int i = 0; i < 20; i++)
							writeBuf[28 + i] = getTrulyRandomByte(); // hash value
						for (int i = 0; i < 20; i++)
							writeBuf[48 + i] = getTrulyRandomByte(); // id value
					} else {
						fillWithTrulyRandomBytes(writeBuf, 0, 68);
					}

					writeLen = 68;
					writePtr = 0;

					log2("Sending handshake message");
					state = STATE_LEECHER_SEND_HANDSHAKE;

					break;
				}
				case STATE_LEECHER_SEND_HANDSHAKE: {
					/* We wait until the entire message has been sent */

					if (writePtr == writeLen) {
						writePtr = writeLen = 0;

						log2("Waiting for handshake message");
						state = STATE_LEECHER_WAIT_HANDSHAKE;
					}

					break;
				}
				case STATE_LEECHER_WAIT_HANDSHAKE: /* Waits for a BitTorrent handshake message */
				{
					/* Now we wait for a handshake message from the Seeder. Once we receive one, we perform some 
					sanity checks. Since we know exactly which programm is running on the other side, this
					check should never fail. If it does, something is seriously wrong! */

					if (readLen >= 68) {
						if (useRealProtocol	&& ((readBuf[0] != 19) || !checkString(readBuf,	1, "BitTorrent protocol"))) {
							log2("Packet is not a BT handshake");
							protocolError = true;
							break;
						}

						log2("Received handshake message");
						chompRead(68);

						/* We respond with a bitfield message. We claim that we have nothing at all. */

						if (useRealProtocol) {
							for (int i = 0; i < 166; i++)
								writeBuf[i] = 0x00;
							
							putInt(writeBuf, 0, 162); // Msg length
							writeBuf[4] = 5;  // msg type
						} else {
							fillWithTrulyRandomBytes(writeBuf, 0, 166);
						}

						writeLen = 166;
						writePtr = 0;

						log2("Sending bitfield message");
						state = STATE_LEECHER_SEND_BITFIELD;
					}

					break;
				}
				case STATE_LEECHER_SEND_BITFIELD: {
					/* We wait until the entire message has been sent */

					if (writePtr == writeLen) {
						writePtr = writeLen = 0;

						log2("Waiting for bitfield message");
						state = STATE_LEECHER_WAIT_BITFIELD;
					}

					break;
				}
				case STATE_LEECHER_WAIT_BITFIELD: {
					/* Now we wait for a bitfield message from the Seeder. */

					if (readLen >= 166) {
						if (useRealProtocol	&& (readBuf[4] != 5)) {
							log2("Packet is not a BT bitfield message");
							protocolError = true;
							break;
						}

						log2("Received bitfield message");
						chompRead(166);

						/* We respond with an 'interested' message. */

						if (useRealProtocol) {
							putInt(writeBuf, 0, 1);
							writeBuf[4] = 2;
						} else {
							fillWithTrulyRandomBytes(writeBuf, 0, 5);
						}

						writeLen = 5;
						writePtr = 0;

						log2("Sending interested message");
						state = STATE_LEECHER_SEND_INTERESTED;
					}

					break;
				}
				case STATE_LEECHER_SEND_INTERESTED: {
					/* We wait until the entire message has been sent */

					if (writePtr == writeLen) {
						writePtr = writeLen = 0;

						log2("Waiting for unchoke message");
						state = STATE_LEECHER_WAIT_UNCHOKE;
					}

					break;
				}
				case STATE_LEECHER_WAIT_UNCHOKE: {
					/* We wait for an 'unchoke' message from the Seeder */

					if (readLen >= 5) {
						if (useRealProtocol	&& (readBuf[4] != 1)) {
							log2("Packet is not a BT unchoke message");
							protocolError = true;
							break;
						}

						log2("Received unchoke message");
						chompRead(5);

						/* We respond with a request message. Note that if we're using random content instead
						of real BitTorrent messages, the Seeder cannot distinguish between a REQUEST and
						a CANCEL. Therefore we set the 12th byte to 1 for REQUESTs, and to 0 for CANCELs.  */						
						if (useRealProtocol) {
							putInt(writeBuf, 0, 13);
							writeBuf[4] = 6;							
							putInt(writeBuf, 5, getTrulyRandomInt() % 1345); // piece index
							putInt(writeBuf, 9, 0); // offset
							putInt(writeBuf, 13, pieceSize); // length
						} else {
							fillWithTrulyRandomBytes(writeBuf, 0, 17);
							writeBuf[11] = 1;
						}

						writeLen = 17;
						writePtr = 0;

						log2("Sending request message");
						state = STATE_LEECHER_SEND_REQUEST;
					}

					break;
				}
				case STATE_LEECHER_SEND_REQUEST: {
					/* We wait until the entire message has been sent */

					if (writePtr == writeLen) {
						writePtr = writeLen = 0;
						statLastIndex = 0;

						//log2("Waiting for piece message");
						state = STATE_LEECHER_WAIT_PIECE;
					}

					break;
				}
				case STATE_LEECHER_WAIT_PIECE: {
					/* Now we're downloading a PIECE message. While the message is not yet complete, we 
					update our statistics in case the conncetion is aborted later. Also, we output
					progress reports periodically. */

					if (readLen > statLastIndex) {
						int newStatBytesTransferred = statBytesTransferred + (readLen - statLastIndex);
						//if (((int) (newStatBytesTransferred / progressReportAfterBytes)) != ((int) (statBytesTransferred / progressReportAfterBytes)))
						//	log2i("Received so far: ", newStatBytesTransferred);

						statBytesTransferred = newStatBytesTransferred;
						statLastIndex = readLen;
					}

					/* Once the message is complete, we parse it */

					if (readLen >= (pieceSize + 13)) {
						if (useRealProtocol && (readBuf[4] != 7)) {
							log2("Packet is not a BT piece message");
							protocolError = true;
							break;
						}

						//log2("Received piece message");
						chompRead(pieceSize + 13);
						statPiecesTransferred++;

						/* And then we send another request message */

						if (useRealProtocol) {
							putInt(writeBuf, 0, 13);
							writeBuf[4] = 6;							
							putInt(writeBuf, 5, getTrulyRandomInt() % 1345); // piece index
							putInt(writeBuf, 9, 0); // offset
							putInt(writeBuf, 13, pieceSize); // length
						} else {
							fillWithTrulyRandomBytes(writeBuf, 0, 17);
							writeBuf[11] = 1;
						}

						writeLen = 17;
						writePtr = 0;

						//log2("Sending request message");
						state = STATE_LEECHER_SEND_REQUEST;
					}

					break;
				}

				/* BitTorrent Seeder ---------------------------------------------------------------------------------*/

				case STATE_SEEDER_WAIT_HANDSHAKE: {
					/* The Seeder initially waits for a handshake message from the Leecher. */

					if (readLen >= 68) {
						if (useRealProtocol	&& ((readBuf[0] != 19) || !checkString(readBuf,	1, "BitTorrent protocol"))) {
							log2("Packet is not a BT handshake");
							protocolError = true;
							break;
						}

						log2("Received handshake message");
						chompRead(68);

						/* We respond with a handshake message of our own */

						if (useRealProtocol) {
							for (int i = 0; i < 68; i++)
								writeBuf[i] = 0;
							writeBuf[0] = 19;
							putString(writeBuf, 1, "BitTorrent protocol");
							for (int i = 0; i < 20; i++)
								writeBuf[28 + i] = readBuf[28 + i]; // the hash the Leecher provided
							for (int i = 0; i < 20; i++)
								writeBuf[48 + i] = getTrulyRandomByte(); // random ID
						} else {
							fillWithTrulyRandomBytes(writeBuf, 0, 68);
						}

						writeLen = 68;
						writePtr = 0;

						log2("Sending handshake message");
						state = STATE_SEEDER_SEND_HANDSHAKE;
					}

					break;
				}
				case STATE_SEEDER_SEND_HANDSHAKE: {
					/* We wait until the entire message has been sent */

					if (writePtr == writeLen) {
						writePtr = writeLen = 0;

						log2("Waiting for bitfield message");
						state = STATE_SEEDER_WAIT_BITFIELD;
					}

					break;
				}
				case STATE_SEEDER_WAIT_BITFIELD: {
					/* We wait for a bitfield message */

					if (readLen >= 166) {
						if (useRealProtocol && (readBuf[4] != 5)) {
							log2("Packet is not a BT bitfield message");
							protocolError = true;
							break;
						}

						log2("Received bitfield message");
						chompRead(166);

						/* We respond with our own bitfield message. Since we're the Seeder, we claim
						that we have everything (bitfield is all ones) */

						if (useRealProtocol) {
							for (int i = 0; i < 166; i++)
								writeBuf[i] = (byte) 0xFF;

							putInt(writeBuf, 0, 162); // Msg length
							writeBuf[4] = 5;  // msg type
						} else {
							fillWithTrulyRandomBytes(writeBuf, 0, 166);
						}

						writeLen = 166;
						writePtr = 0;

						log2("Sending bitfield message");
						state = STATE_SEEDER_SEND_BITFIELD;
					}

					break;
				}
				case STATE_SEEDER_SEND_BITFIELD: {
					/* We wait until the entire message has been sent */

					if (writePtr == writeLen) {
						writePtr = writeLen = 0;

						log2("Waiting for interested message");
						state = STATE_SEEDER_WAIT_INTERESTED;
					}

					break;
				}
				case STATE_SEEDER_WAIT_INTERESTED: {
					/* We wait for an 'interested' message from the Leecher */

					if (readLen >= 5) {
						if (useRealProtocol	&& (readBuf[4] != 2)) {
							log2("Packet is not a BT interested message");
							protocolError = true;
							break;
						}

						log2("Received interested message");
						chompRead(5);

						/* We respond with an unchoke message */

						if (useRealProtocol) {
							putInt(writeBuf, 0, 1);
							writeBuf[4] = 1;							
						} else {
							fillWithTrulyRandomBytes(writeBuf, 0, 5);
						}

						writeLen = 5;
						writePtr = 0;

						log2("Sending unchoke message");
						state = STATE_SEEDER_SEND_UNCHOKE;
					}

					break;
				}
				case STATE_SEEDER_SEND_UNCHOKE: {
					/* We wait until the entire message has been sent */

					if (writePtr == writeLen) {
						writePtr = writeLen = 0;

						log2("Waiting for request or cancel message");
						state = STATE_SEEDER_WAIT_REQUEST_OR_CANCEL;
					}

					break;
				}
				case STATE_SEEDER_WAIT_REQUEST_OR_CANCEL: {
					/* At this point, the next message could either be a REQUEST or a CANCEL.
					If we're using the randomized protocol, we distinguish between the two
					using the 12th byte. */
					
					if ((readLen >= 17)	&& (useRealProtocol ? (readBuf[4] == 8) : (readBuf[11] == 0))) {
						log2("Received cancel message");
						connectionClosed = true;
					} else if ((readLen >= 17) && (useRealProtocol ? (readBuf[4] == 6) : (readBuf[11] == 1))) {
						//log2("Received request message");
						chompRead(17);

						if (useRealProtocol) {				
						
							/* Extract parameters from the REQUEST message */

							int blength = getInt(readBuf, 13);
							int index = getInt(readBuf, 5);
							int offset = getInt(readBuf, 9);

							/* Perform some sanity checks */

							if (blength > pieceSize)
								blength = pieceSize;
							if (blength < 0)
								blength = pieceSize;

							/* Make a piece message. */

							putInt(writeBuf, 0, 9 + pieceSize);
							writeBuf[4] = 7;
							putInt(writeBuf, 5, index);
							putInt(writeBuf, 9, offset);

							fillWithPseudoRandomBytes(writeBuf, 13, blength);
							writeLen = blength + 13;
						}
						else {
							fillWithPseudoRandomBytes(writeBuf, 0, pieceSize + 13);
							writeLen = pieceSize + 13;
						}							
						
						writePtr = 0;
						statLastIndex = 13;

						//log2("Sending piece message");
						state = STATE_SEEDER_SEND_PIECE;
					} else if (readLen >= 17) {
						log2("Packet is neither a BT request nor a BT cancel");
						protocolError = true;
						break;
					}

					break;
				}
				case STATE_SEEDER_SEND_PIECE: {
					/* We wait until the entire message has been sent. Again, we keep our statistics updated
					in case the connection is suddenly aborted. */

					if (writePtr > statLastIndex) {
						int newStatBytesTransferred = statBytesTransferred + (writePtr - statLastIndex);
						//if (((int) (newStatBytesTransferred / progressReportAfterBytes)) != ((int) (statBytesTransferred / progressReportAfterBytes)))
						//	log2i("Sent so far: ", newStatBytesTransferred);

						statBytesTransferred = newStatBytesTransferred;
						statLastIndex = writePtr;
					}

					if (writePtr == writeLen) {
						writePtr = writeLen = 0;
						statPiecesTransferred++;

						//log2("Waiting for next request or cancel message");
						state = STATE_SEEDER_WAIT_REQUEST_OR_CANCEL;
					}

					break;
				}
				default: {
					log2("Unknown protocol state");
					protocolError = true;
					break;
				}
				}

				/* End of the BitTorrent state machine. The code below is no longer portable.                  *
				/***********************************************************************************************/

				if (protocolError || connectionClosed)
					break;

				long now = System.currentTimeMillis();
				if (now >= endTime){
					log2("Timeout. Transfer done (" + now + ">=" + endTime + ")");
					break;
				}

				/* We use the NIO select() call. */

				long selTimeout = endTime - now;
				int numReady = -1;
				try {
					numReady = selector.select(selTimeout);
				} catch (IOException e) {
					System.err.println("Cannot call select: " + e);
					e.printStackTrace();
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}

					handleFatalError("Internal error", "id=" + id	+ "&msg=select%20failed", 4);
					transferSuccessful = false;
					return statBytesTransferred;
				}

				/* Check if any keys are ready, and if so, iterate over them */

				if (numReady > 0) {
					Iterator it = selector.selectedKeys().iterator();
					while (it.hasNext()) {
						SelectionKey selKey = (SelectionKey) it.next();
						it.remove();

						if (!selKey.isValid())
						    continue;

						/* If the channel is ready for reading, do so */
						
						if (selKey.isReadable()) {
							sChannel = (SocketChannel) selKey.channel();

							try {
								rbuf.limit(readBufSizeBytes);
								rbuf.position(readLen);
								int bytesRead = sChannel.read(rbuf);

								if (bytesRead <= 0) {
									log2("Read returned " + bytesRead + ". Transfer finished.");
									connectionClosed = true;
									break;
								}

								readLen += bytesRead;
							} catch (IOException e) {
								// The channel was closed...
								backLog += "Exception: " + e.getMessage()	+ " (state="+state+")\n";
								System.err.println("Exception: " + e.getMessage());
								e.printStackTrace();

								/* Here we run into a problem: The only way to distinguish between connection 
								resets and other problems is to parse the text of the exception. 
								Unfortunately some JVMs localize their exception messages, so for all we know
								this might be in Chinese. We check for English immediately, but we report the
								exception message to the server just in case, so we can do postprocessing. */

								if (e.getMessage().contains("reset") || e.getMessage().contains("forcibly")) {
									System.err.println("Connection was resetted.");
									resetted = true;
								} else {
									System.err.println("Connection failed: " + e.getMessage());
								}

								selKey.cancel();
								rbuf.clear();
								wbuf.clear();
								connectionClosed = true;
							}
						}

						/* If the channel is ready for writing, do so */

						if (!connectionClosed && selKey.isWritable() && (writePtr < writeLen)) {
							sChannel = (SocketChannel) selKey.channel();

							try {
								wbuf.limit(writeLen);
								wbuf.position(writePtr);
								int bytesWritten = sChannel.write(wbuf);

								if (bytesWritten <= 0) {
									log2("Write returned " + bytesWritten + ". Transfer finished.");
									connectionClosed = true;
									break;
								}

								writePtr += bytesWritten;
							} catch (IOException e) {
								// The channel was closed...
								backLog += "Exception: " + e.getMessage()	+ " (state="+state+")\n";
								System.err.println("Exception: " + e.getMessage());
								e.printStackTrace();

								/* Again, we have a problem with localized exception texts (see comment above) */

								if (e.getMessage().contains("reset") || e.getMessage().contains("forcibly")) {
									System.err.println("Connection was resetted.");
									resetted = true;
								} else {
									System.err.println("Connection failed: " + e.getMessage());
								}

								selKey.cancel();
								rbuf.clear();
								wbuf.clear();
								connectionClosed = true;
							}
						}
					}
				} else {
					connectionClosed = true;
					break;
				}
			}

			//System.err.println("End of transfer; " + statBytesTransferred + " bytes and " + statPiecesTransferred + " pieces transferred");

			transferSuccessful = (!protocolError && (statBytesTransferred > 10000));

			return statBytesTransferred;
		}

		/**
		 * Acts as a tcp receiver
		 * @param sChannel A socket connected to the receiver
		 * @param selector The selector the socket is registered with
		 * @param timeout The maximum transfer time
		 * @return The number of bytes transfered.
		 */
		public double tcpReceiver(SocketChannel sChannel, Selector selector, long timeout) {

			boolean isTerminated = false;
			transferSuccessful = true;
			resetted = false;
			long transferEnd = System.currentTimeMillis() + timeout;
			ByteBuffer buf = ByteBuffer.allocate(32000);

			double receivedBytes = 0;
			while ((!isTerminated)) {
				timeout = transferEnd - System.currentTimeMillis();
				if (timeout <= 0)
					return receivedBytes;
				int numReady = -1;
				try {
					numReady = selector.select(timeout);
				} catch (IOException e) {
					System.err.println("Cannot call select: " + e);
					e.printStackTrace();
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}
					handleFatalError("Internal error", "id=" + id	+ "&msg=select%20failed", 4);
					transferSuccessful = false;
					return receivedBytes;
				}

				if (numReady > 0) {
					Iterator it = selector.selectedKeys().iterator();
					while (it.hasNext()) {
						SelectionKey selKey = (SelectionKey) it.next();
						it.remove();
						buf.clear();

						if (selKey.isReadable()) {
							try {
								int bytesRead = sChannel.read(buf);

								if (bytesRead <= 0) {
									isTerminated = true;
									break;
								}

								receivedBytes += bytesRead;

							} catch (IOException e) {
								// The channel was closed...
								//if(e.getMessage().equals("Connection reset by peer")){
								backLog += "tcpReceiver: " + e.getMessage()	+ "\n";
								if (e.getMessage().contains("reset") || e.getMessage().contains("forcibly")) {
									System.err.println("Connection was resetted.");
									resetted = true;
								} else {
									System.err.println("Connection failed: " + e.getMessage());
								}
								selKey.cancel();

								buf.clear();
								isTerminated = true;
								if (receivedBytes <= 0)
									transferSuccessful = false;
								break;
							}
						} else if (selKey.isWritable()) {
						}
					}
				} else {
					// Timeout
					isTerminated = true;
					break;
				}
			}
			return receivedBytes;
		}

		/**
		 * Acts as a TCP sender
		 * @param sChannel A socket connected to the receiver
		 * @param selector The selector the socket is registered with
		 * @param timeout The maximum transfer time
		 * @return The number of bytes transfered.
		 */

		public double tcpSender(SocketChannel sChannel, Selector selector, long timeout) {

			boolean isTerminated = false;
			transferSuccessful = true;
			resetted = false;
			long transferEnd = System.currentTimeMillis() + timeout;
			ByteBuffer buf = ByteBuffer.allocate(16000);
			Random rnd = new Random();

			double sentBytesTotal = 0;
			while ((!isTerminated)) {
				timeout = transferEnd - System.currentTimeMillis();
				if (timeout <= 0)
					return sentBytesTotal;
				int numReady = -1;
				try {
					numReady = selector.select(timeout);
				} catch (IOException e) {
					System.err.println("Cannot call select: " + e);
					e.printStackTrace();
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}
					handleFatalError("Internal error", "id=" + id	+ "&msg=select%20failed", 4);
					transferSuccessful = false;
					return sentBytesTotal;
				}

				if (numReady > 0) {
					Iterator it = selector.selectedKeys().iterator();
					while (it.hasNext()) {
						SelectionKey selKey = (SelectionKey) it.next();
						it.remove();
						buf.clear();

						if (selKey.isReadable()) {
						} else if (selKey.isWritable()) {
							rnd.nextBytes(buf.array());
							buf.position(buf.limit());
							buf.flip();
							try {
								int sentBytes = sChannel.write(buf);
								if (sentBytes <= 0) {
									isTerminated = true;
									break;
								}
								sentBytesTotal += sentBytes;
							} catch (IOException e) {
								// The channel was closed...
								//if(e.getMessage().equals("Connection reset by peer")){
								backLog += "tcpSender: " + e.getMessage() + "\n";
								if (e.getMessage().contains("reset") || e.getMessage().contains("forcibly")) {
									System.err.println("Connection was resetted.");
									resetted = true;
								} else {
									System.err.println("Connection failed: " + e.getMessage());
								}
								selKey.cancel();

								buf.clear();
								isTerminated = true;
								if (sentBytesTotal <= 0)
									transferSuccessful = false;
								break;
							}
						}
					}
				} else {
					// Timeout
					isTerminated = true;
					break;
				}
			}
			return sentBytesTotal;
		}

		/**
		 * Create the command string that's sent to the server/peer
		 */
		String getCommandStr(boolean setupUpstream, boolean setupDownstream, int btUpstream, int tcpUpstream, int btDownstream, 	int tcpDownstream) {
			String command = null;
			int port = -1;
			if ((server2 != null) && (btUpstream <= 0) && (tcpUpstream <= 0) && (btDownstream <= numberOfRepeats) && (tcpDownstream <= numberOfRepeats))
				port = server2.getPort();
			else if ((server2 != null) && (btUpstream <= numberOfRepeats) && (tcpUpstream <= numberOfRepeats) && ((btUpstream > 0) || (tcpUpstream > 0)))
				port = server2.getPort();
			else
				port = server.getPort();

			// Command string: "<ExperimentName> <Direction> <Duration> port <PortNumber> .\n"
			if (setupUpstream) {
				command = "setup downstream " + ((setupDuration / 1000) + 1) + " port " + port + " .\n";
			} else if ((server2 != null) && (port == server2.getPort())	&& setupUpstream2) {
				command = "setup downstream " + ((setupDuration / 1000) + 1) + " port " + port + " .\n";
			} else if ((btUpstream > 0) && (btUpstream >= tcpUpstream)) {
				command = "bt downstream " + (int) (1 + (duration / 1000)) + " port " + port + " .\n";
			} else if (tcpUpstream > 0) {
				command = "tcp downstream " + (int) (1 + (duration / 1000)) + " port " + port + " .\n";
			} else if (setupDownstream) {
				command = "setup upstream " + (int) (setupDuration / 1000) + " port " + port + "\n";
			} else if ((server2 != null) && (port == server2.getPort()) && setupDownstream2) {
				command = "setup upstream " + (int) (setupDuration / 1000) + " port " + port + " .\n";
			} else if ((btDownstream > 0) && (btDownstream >= tcpDownstream)) {
				command = "bt upstream " + (int) (duration / 1000) + " port " + port + " .\n";

			} else if (tcpDownstream > 0) {
				command = "tcp upstream " + (int) (duration / 1000) + " port " + port + " .\n";
			}

			return command;
		}

		public void run() {

			int soTimeout = 10000; // 10 seconds

			Selector selector = null;
			SocketChannel sChannel = null;
			SelectionKey selKey = null;

			/* Java on MacOS X prevents applets from opening a selector due to some property reading that's not allowed.
			 * Thus, the applet must be signed and the code below allows us to escape the sandbox.
                         *
                         * NOTE: The Makefile automatically builds JARs with and without this code. It uses the MACONLY
                         * markers to uncomment the code, so please do not remove these markers.
			 */
                         
/* MACONLY
			if(btc != null){
			selector = (Selector) AccessController.doPrivileged(new PrivilegedAction() {
				public Object run() {
					Selector selector = null;
					try {
						selector = Selector.open();
					} catch (IOException e1) {
						System.err.println("Creating selector failed." + e1);
						e1.printStackTrace();
						String id = null;
						if(btc != null){
							id = btc.userid;
						}
						handleFatalError("Internal error", "id=" + id + "&msg=selector%20failed", 4);
						return null;
					}

					return selector;
				}
			});
			if(selector == null)
				return;
			}
			else{
MACONLY */
			try {
				selector = Selector.open();
			} catch (IOException e1) {
				System.err.println("Creating selector failed." + e1);
				e1.printStackTrace();
				String id = null;
				if (btc != null) {
					id = btc.userid;
				}
				handleFatalError("Internal error", "id=" + id + "&msg=selector%20failed", 4);
				return;
			}
/* MACONLY
                        }
MACONLY */

			// Open a TCP connection to the server
			double bytesTransmitted = 0;

			SocketChannel commandChannel = null;
			ByteBuffer command = ByteBuffer.allocate(32000);
			try {
				commandChannel = SocketChannel.open(new InetSocketAddress(server.getAddress(), commandPort));
				commandChannel.configureBlocking(true);
				commandChannel.socket().setSoTimeout(soTimeout);
			} catch (IOException e) {
				System.err.println("Cannot connect to server.");
				String id = null;
				if (btc != null) {
					id = btc.userid;
				}
				handleFatalError("Cannot connect to server", "id=" + id + "&msg=Cannot%20connect%20to%20server&mid=10", 8, e);
				return;
			}

			if(!sysinfo.equals(""))
				backLog += sysinfo + '\n';
			
			String results = "";
			int btup = 0, btdown = 0, tcpup = 0, tcpdown = 0;

			while (!isTerminated && (setupUpstream || setupDownstream || ((btUpstream + tcpUpstream + btDownstream + tcpDownstream) > 0))) {
				bytesTransmitted = 0;

				// Transmit over command socket what to do next and wait for answer
				String commandStr = getCommandStr(setupUpstream, setupDownstream, btUpstream, tcpUpstream, btDownstream, tcpDownstream);
				if (commandStr == null) {
					//System.err.println("FATAL: Do not know what experiment to run next. Aborting.");
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}
					handleFatalError("Setup failed", "id=" + id + "&msg=No%20command", 30);
					return;
				}
				try {
					command.put(commandStr.getBytes("US-ASCII"));
				} catch (UnsupportedEncodingException e2) {
					System.exit(99);
				}
				command.flip();
				try {
					//System.out.print("Sending command: " + commandStr);
					commandChannel.write(command);
				} catch (IOException e2) {
					//System.err.println("Cannot send command to server.");
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}
					handleFatalError("Server failed", "id=" + id + "&msg=Server%20failed&mid=11", 31, e2);
					return;
				}
				command.clear();
				try {
					int ret = 0;
					while (ret == 0) {
						ret = commandChannel.read(command);

						if (ret < 0) {
							//System.err.println("Cannot read from command socket");
							String id = null;
							if (btc != null) {
								id = btc.userid;
							}
							handleFatalError("Server closed connection prematurely", "id=" + id + "&msg=Server%20connection%20closed%20prematurely", 32);
							return;
						}
					}

					boolean isOK = true;
					byte[] cs = "ok\n".getBytes("US-ASCII");
					for (int i = 0; i < cs.length; i++) {
						if (cs[i] != command.array()[i]) {
							isOK = false;
							break;
						}
					}
					command.clear();
					if (!isOK) {
						System.err.println("Unknown response from server.");
						continue;
					} else
						System.out.println("Server responded with OK.");
				} catch (SocketTimeoutException e) {
					//System.err.println("Server did not respond to our command. Aborting.");
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}
					handleFatalError("Server timeout", "id=" + id	+ "&msg=Server%20not%20responding&mid=11", 33, e);
					return;
				} catch (IOException e) {
					//System.err.println("Cannot read from command socket.");
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}
					handleFatalError("Server connection IO failed",	"id="	+ id + "&msg=Server%20connection%20failed&mid=11", 34, e);
					return;
				}

				if ((server2 != null) && (btUpstream <= 0) && (tcpUpstream <= 0) && (btDownstream <= numberOfRepeats) && (tcpDownstream <= numberOfRepeats))
					sChannel = setupSocket(server2, soTimeout);
				else if ((server2 != null) && (btUpstream <= numberOfRepeats) && (tcpUpstream <= numberOfRepeats) && ((btUpstream > 0) || (tcpUpstream > 0)))
					sChannel = setupSocket(server2, soTimeout);
				else
					sChannel = setupSocket(server, soTimeout);

				if (sChannel == null) {
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}
					handleFatalError("Cannot create socket", "id=" + id + "&msg=Cannot%20connect%20socket&mid=10", 22);
					return;
				}

				try {
					selKey = sChannel.register(selector, SelectionKey.OP_WRITE | SelectionKey.OP_READ);
				} catch (ClosedChannelException e1) {
					System.err.println("Cannot register socket with select: " + e1);
					e1.printStackTrace();
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}
					handleFatalError("Internal error", "id=" + id + "&msg=register%20failed", 8, e1);
					return;
				}

				// This will first execute all upstream experiments and then downstream experiments
				if (setupUpstream) {
					System.out.println("Setting up upstream transfer.");
					if (cd != null)
						cd.addToBound(10);
					tcpSender(sChannel, selector, setupDuration);
					setupUpstream = false;
				} else if ((server2 != null) && (sChannel.socket().getPort() == server2.getPort()) && setupUpstream2) {
					System.out.println("Setting up upstream transfer2.");
					if (cd != null)
						cd.addToBound(10);
					tcpSender(sChannel, selector, setupDuration);
					setupUpstream2 = false;
				} else if ((btUpstream > 0) && (btUpstream >= tcpUpstream)) {
					System.out.println("Running BT upstream transfer.");
					if (cd != null)
						cd.addToBound((int) (duration / 1000.0));
					start = System.currentTimeMillis();
					bytesTransmitted = connectionHandler(sChannel, selector,
							duration, true, true);
					end = System.currentTimeMillis();
					btUpstream--;

					double bps = 0;
					if ((end - start) > 0)
						bps = bytesTransmitted * 8000.0 / (end - start);

					if (resetted && (bytesTransmitted == 0)) {
						results += "btup" + btup + "=reset&";
						System.out.println("FAILED: BT upstream transfer resetted.");
						backLog += "FAILED: BT upstream transfer resetted.\n";
					} else if (resetted) {
						results += "btup" + btup + "=" + (long) bps + "r&";
						System.out.println("FAILED: BT upstream transfer resetted after a while.");
						backLog += "FAILED: BT upstream transfer resetted after a while.\n";
					} else {
						results += "btup" + btup + "=" + (long) bps + "&";
					}
					System.out.println("Transfered " + bytesTransmitted	+ " bytes in " + ((end - start) / 1000.0)	+ " seconds: " + bps + " bps");
					backLog += "Transfered " + bytesTransmitted + " bytes in " + ((end - start) / 1000.0) + " seconds: " + bps + " bps\n";
					btup++;
				} else if (tcpUpstream > 0) {
					System.out.println("Running TCP upstream transfer.");
					if (cd != null)
						cd.addToBound((int) (duration / 1000.0));
					start = System.currentTimeMillis();
					bytesTransmitted = connectionHandler(sChannel, selector, duration, false, true);
					end = System.currentTimeMillis();
					tcpUpstream--;

					double bps = 0;
					if ((end - start) > 0)
						bps = bytesTransmitted * 8000.0 / (end - start);

					if (resetted && (bytesTransmitted == 0)) {
						results += "tcpup" + tcpup + "=reset&";
						System.out.println("FAILED: TCP upstream transfer resetted.");
						backLog += "FAILED: TCP upstream transfer resetted.\n";
					} else if (resetted) {
						results += "tcpup" + tcpup + "=" + (long) bps + "r&";
						System.out.println("FAILED: TCP upstream transfer resetted after a while.");
						backLog += "FAILED: TCP upstream transfer resetted after a while.\n";
					} else {
						results += "tcpup" + tcpup + "=" + (long) bps + "&";
					}
					System.out.println("Transfered " + bytesTransmitted	+ " bytes in " + ((end - start) / 1000.0)	+ " seconds: " + bps + " bps");
					backLog += "Transfered " + bytesTransmitted + " bytes in " + ((end - start) / 1000.0) + " seconds: " + bps + " bps\n";
					tcpup++;
				} else if (setupDownstream) {
					System.out.println("Setting up downstream transfer.");
					if (cd != null)
						cd.addToBound(10);
					tcpReceiver(sChannel, selector, setupDuration + 1000);
					setupDownstream = false;
				} else if ((server2 != null) && (sChannel.socket().getPort() == server2.getPort()) && setupDownstream2) {
					System.out.println("Setting up downstream transfer2.");
					if (cd != null)
						cd.addToBound(10);
					tcpSender(sChannel, selector, setupDuration + 1000);
					setupDownstream2 = false;
				} else if ((btDownstream > 0) && (btDownstream >= tcpDownstream)) {
					System.out.println("Running BT downstream transfer.");
					if (cd != null)
						cd.addToBound((int) (duration / 1000.0));
					start = System.currentTimeMillis();
					bytesTransmitted = connectionHandler(sChannel, selector, duration, true, false);
					end = System.currentTimeMillis();
					btDownstream--;
					double bps = 0;
					if ((end - start) > 0)
						bps = bytesTransmitted * 8000.0 / (end - start);

					if (resetted && (bytesTransmitted == 0)) {
						results += "btdown" + btdown + "=reset&";
						System.out.println("FAILED: BT downstream transfer resetted.");
						backLog += "FAILED: BT downstream transfer resetted.\n";
					} else if (resetted) {
						results += "btdown" + btdown + "=" + (long) bps + "r&";
						System.out.println("FAILED: BT downstream transfer resetted after a while.");
						backLog += "FAILED: BT downstream transfer resetted after a while.\n";
					} else {
						results += "btdown" + btdown + "=" + (long) bps + "&";
					}
					System.out.println("Transfered " + bytesTransmitted	+ " bytes in " + ((end - start) / 1000.0)	+ " seconds: " + bps + " bps");
					backLog += "Transfered " + bytesTransmitted + " bytes in " + ((end - start) / 1000.0) + " seconds: " + bps + " bps\n";
					btdown++;
				} else if (tcpDownstream > 0) {
					System.out.println("Running TCP downstream transfer.");
					if (cd != null)
						cd.addToBound((int) (duration / 1000.0));
					start = System.currentTimeMillis();
					bytesTransmitted = connectionHandler(sChannel, selector, duration, false, false);
					end = System.currentTimeMillis();
					tcpDownstream--;

					double bps = 0;
					if ((end - start) > 0)
						bps = bytesTransmitted * 8000.0 / (end - start);

					if (resetted && (bytesTransmitted == 0)) {
						results += "tcpdown" + tcpdown + "=reset&";
						System.out.println("FAILED: TCP downstream transfer resetted.");
						backLog += "FAILED: TCP downstream transfer resetted.\n";
					} else if (resetted) {
						results += "tcpdown" + tcpdown + "=" + (long) bps + "r&";
						System.out.println("FAILED: TCP downstream transfer resetted.");
						backLog += "FAILED: TCP downstream transfer resetted after a while.\n";
					} else {
						results += "tcpdown" + tcpdown + "=" + (long) bps + "&";
					}
					System.out.println("Transfered " + bytesTransmitted	+ " bytes in " + ((end - start) / 1000.0)	+ " seconds: " + bps + " bps");
					tcpdown++;
					backLog += "Transfered " + bytesTransmitted + " bytes in " + ((end - start) / 1000.0) + " seconds: " + bps + " bps\n";
				} else {
					break;
				}

				if (backLog.length() > 0) {

					String part[] = backLog.split("\n");

					try {
						for (int i = 0; i < part.length; i++) {
							if (part[i].length() == 0)
								continue;

							command.put("log ".getBytes("US-ASCII"));
							command.put(part[i].getBytes("US-ASCII"));
							command.put("\n".getBytes("US-ASCII"));
						}
					} catch (UnsupportedEncodingException e2) {
						System.exit(99);
					}
					command.flip();
					try {
						//System.out.print("Sending log: " + backLog);
						commandChannel.write(command);
					} catch (IOException e2) {
						//System.err.println("Cannot send log to server.");
						String id = null;
						if (btc != null) {
							id = btc.userid;
						}
						handleFatalError("Server failed", "id=" + id + "&msg=Server%20failed&mid=11", 31, e2);
						return;
					}
					command.clear();
					backLog = "";
				}

				// Clean up
				try {
					selKey.cancel();
					if (sChannel.isConnected()) {
						//sChannel.configureBlocking(true);
						sChannel.socket().close();
					}
				} catch (IOException e) {
					//System.err.println("Cannot close socket. " + e);
					String id = null;
					if (btc != null) {
						id = btc.userid;
					}
					handleFatalError("Socket failure", "id=" + id + "&msg=Cannot%20close%20socket&mid=11", 36, e);
					return;
				}

				// Sleep X seconds and give the server some time to prepare for the next round and to drain queues
				try {
					sleep(pause);
				} catch (InterruptedException e) {
				}
			}

			try {
				selector.close();
			} catch (IOException e) {
				System.err.println("Cannot close selector.");
			}

			// Transmit over command socket what to do next and wait for answer
			String commandStr = "shutdown \n";
			try {
				command.put(commandStr.getBytes("US-ASCII"));
			} catch (UnsupportedEncodingException e2) {
				System.exit(99);
			}
			command.flip();
			try {
				commandChannel.write(command);
				command.clear();
			} catch (IOException e2) {
				System.err.println("Cannot send command to server.");
				// return; // Let's ignore this problem for the moment
			}

			// Wait for a final OK, giving the server time to finish its business
			try {
				int ret = 0;
				while (ret == 0) {
					ret = commandChannel.read(command);

					if (ret < 0) {
						//System.err.println("Cannot read from command socket");
						break;
					}
				}

				boolean isOK = true;
				byte[] cs = "ok\n".getBytes("US-ASCII");
				for (int i = 0; i < cs.length; i++) {
					if (cs[i] != command.array()[i]) {
						isOK = false;
						break;
					}
				}
				command.clear();
				if (!isOK) {
					System.err.println("Unknown response from server.");
				} else{
					System.out.println("Server responded with OK.");
				}
				
				// Finally close the command channel
				commandChannel.close();
			} catch (Exception e) {} // Ignore at this point			
			
			
			// We are done, display results
			if (btc != null) {
				btc.displayResultPage(results);
				btc.finished = true;
				btc.btw = null;
			} else {
				System.out.println(results);
			}

		}

		private void handleFatalError(String message, String errorParams, int errorCode) {
			this.handleFatalError(message, errorParams, errorCode, null);
		}

		private void handleFatalError(String message, String errorParams, int errorCode, Exception e) {

			isTerminated = true;

			if ((btc != null) && (errorParams != null)) {

				if (btc.btw != null)
					btc.btw.isTerminated = true;
				btc.printItem(message);

				URL nextPage;
				try {
					// Encode text
					String[] part;
					if (!sysinfo.equals("")) {
						part = sysinfo.split(" ");
						sysinfo = "sysinfo=" + part[0];
						for (int i = 1; i < part.length; i++) {
							sysinfo += "%20" + part[i];
						}
					}
					String exception = "";
					if (e != null) {
						part = e.getMessage().split(" ");
						exception = "exception=" + part[0];
						for (int i = 1; i < part.length; i++) {
							exception += "%20" + part[i];
						}
					}

					String nextPageStr = "http://"	+ btc.PlServer.getAddress().getHostAddress() + "/" + btc.errorpage + "?"+errorParams + "&" + exception + "&" + sysinfo;
					try{
						nextPage = new URL(nextPageStr);
						btc.displayResultPage(nextPage);  // Go to a special error handling webpage
					} catch (MalformedURLException me) {
						System.err.println("URL '" + nextPageStr + "' malformed: " + me);
						e.printStackTrace();
					}
				} catch (Exception me) {
					me.printStackTrace();
				}
			} else {
				System.err.println("Error: " + message);
				if (e != null) {
					System.err.println("Exception: " + e.getMessage());
					e.printStackTrace();
				}
				if (!sysinfo.equals("")) {
					System.err.println("Sysinfo: " + sysinfo);
				}
				System.exit(errorCode);
			}
		}
	}

	/** Initializes the applet */

	public void init() {

		buffer = new StringBuffer();

		String port = getParameter("port");
		if (port == null) {
			port = "6881";
		}

		String port2 = getParameter("port2");

		String str = getParameter("down");
		if (str != null && str.equalsIgnoreCase("true"))
			this.down = true;
		str = getParameter("up");
		if (str != null && str.equalsIgnoreCase("true"))
			this.up = true;
		str = getParameter("tcp");
		if (str != null && str.equalsIgnoreCase("true"))
			this.tcp = true;
		str = getParameter("repeat");
		if (str != null)
			this.repeat = Integer.valueOf(str);
		str = getParameter("duration");
		if (str != null)
			this.duration = Integer.valueOf(str);

		if (!up && !down) {
			System.err.println("Misconfiguration: Choose either or both up and down traffic!");
			return;
		}
		
		str = getParameter("errorpage"); // The page to call when there's an error
		if(str != null)
			errorpage = str;
		str = getParameter("testpage"); // The page to call when the measurement is finished
		if(str !=  null)
			testpage = str;
		
		// The planet just can be the server we got this applet from
		String PLANET = getCodeBase().getHost();
		//		String PLANET = "loki03.mpi-sws.mpg.de"; // Eclipse debug
		PlServer = new InetSocketAddress(PLANET, Integer.valueOf(port));
		if ((port2 != null) && (Integer.valueOf(port2) > 0))
			PlServer2 = new InetSocketAddress(PLANET, Integer.valueOf(port2));

		String id = getParameter("ID");
		if (id != null) {
			this.userid = id;
		}

		setBackground(Color.WHITE);
	}

	public void start() {
		printItem("Starting measurement...");

		sysinfo = System.getProperty("os.name") + ',' + System.getProperty("os.arch") + ',' + System.getProperty("os.version") + ',' + System.getProperty("java.vendor") + ',' + System.getProperty("java.version");

		// Only run one instance of the Worker thread!
		if (btw == null) {
			btw = new BitTorrentWorker(PlServer, this);
			btw.confExp(up, down, tcp, repeat, PlServer2, duration);
			btw.start();

			appletContext = getAppletContext();
		}
	}

	public void stop() {
		//		printItem("Stopping applet");
		//		btw.isTerminated = true; // ??? Is this called whenever the applet looses the focus?
	}

	public void destroy() {

		if (btw != null) {
			btw.isTerminated = true;
			btw = null;
		}
	}

	public void displayResultPage(String results) {
		/* Finally, open the web page that shows the results */
		String nextURL = "http://" + PlServer.getAddress().getHostAddress() + '/' + testpage + "?id=" + userid + "&done=yes&btport=" + PlServer.getPort() + "&repeat=" + repeat + '&';
		if (PlServer2 != null)
			nextURL += "port2=" + PlServer2.getPort() + '&';
		
		// Encode text
		String[] part;
		if (!sysinfo.equals("")) {
			part = sysinfo.split(" ");
			sysinfo = "sysinfo=" + part[0];
			for (int i = 1; i < part.length; i++) {
				sysinfo += "%20" + part[i];
			}
			
			nextURL += sysinfo + '&';
		}
		
		nextURL += results;
		
		printItem("Opening results page...");
		URL url = null;
		try {
			url = new URL(nextURL);
		} catch (MalformedURLException e) {
			System.err.println("Malformed URL: " + nextURL);
		}

		if (url != null) {
			appletContext.showDocument(url);
		}
	}

	public void displayResultPage(URL url) {

		if (url != null) {
			printItem("Opening next page...");
			appletContext.showDocument(url);
		}
	}

	private void addItem(String newWord) {
		buffer.append(newWord);
		repaint();
	}

	private void printItem(String newWord) {
		buffer.delete(0, buffer.length());
		buffer.append(newWord);
		repaint();
	}

	/**
	 * Plots to the applet frame text and does also a progress bar.
	 */

	public void paint(Graphics g) {
		Dimension FrameDimension = getSize();
		int BarPixelWidth = (FrameDimension.width * done) / totalLength;

		// Fill the bar the appropriate percent full.
		g.setColor(Color.decode("0x69acff")); // MPI blue

		//Draw a Rectangle around the applet's display area.
		g.drawRect(0, 0, getWidth() - 1, getHeight() - 1);

		// Draw progress bar
		g.fillRect(0, 0, BarPixelWidth, FrameDimension.height);

		// Set the color of the text
		g.setColor(Color.black);

		// Calculate the width of the string in pixels. Used to center the string in the progress bar window
		FontMetrics fm = g.getFontMetrics(g.getFont());
		int StringPixelWidth = fm.stringWidth(buffer.toString());

		g.drawString(buffer.toString(),
				(FrameDimension.width - StringPixelWidth) / 2, 24);
	}

	public String[][] getParameterInfo() {
		String[][] info = {
				// Parameter Name     Kind of Value   Description
				{ "ID", "String", "The unique ID of this host" },
				{ "port", "Integer", "Port of the server to connect to (default: 6881)" },
				{ "port2", "Integer", "Another port of the server to connect to" },
				{ "down", "Boolean", "Whether to emulate downstream traffic (true or false)" },
				{ "up", "Boolean", "Whether to emulate upstream traffic (true or false)" },
				{ "tcp", "Boolean", "Whether to emulate TCP traffic (true or false)" },
				{ "repeat", "Integer", "How often a single experiment should be repeated" },
				{ "duration", "Integer", "The length of a single test (in seconds, default: 20)" }, 
			};
		return info;
	}

	public String getAppletInfo() {
		return "BitTorrenClient v2.1\nEmulates a BitTorrent transfer.\nContact: Marcel Dischinger <mdischin@mpi-sws.mpg.de>\n";
	}

	/**
	 * What parameter main() takes.
	 * nodeID is for compatibility with the SatLab test harness. It's value is not used. 
	 */
	public static void usage() {
		System.err.println("Usage: BitTorrentClient <nodeID> <serverIP> <BitTorrentPort> [-up (true|false)] [-down (true|false)] [-tcp] [-p2 <port2>] [-r <number>] [-t <DurationInSeconds>] [-o <outputFile>]");
		System.exit(1);
	}

	/**
	 * @param args see usage()
	 */
	public static void main(String[] args) {

		if (args.length < 3 || args.length > 16) {
			System.err.println("Error: Too many parameters!");
			usage();
		}

		int port2 = -1;
		boolean tcp = false, up = false, down = false;
		int repeat = 1;
		int duration = -1;
		String filename = null;

		for (int i = 3; i < args.length; i++) {

			if (args[i].equalsIgnoreCase("-p2")
					|| args[i].equalsIgnoreCase("-port2")) {
				i++;
				if (i >= args.length)
					usage();
				port2 = Integer.valueOf(args[i]);
			} else if (args[i].equalsIgnoreCase("-up")) {
				i++;
				if (i >= args.length)
					usage();
				up = Boolean.valueOf(args[i]);
			} else if (args[i].equalsIgnoreCase("-down")) {
				i++;
				if (i >= args.length)
					usage();
				down = Boolean.valueOf(args[i]);
			} else if (args[i].equalsIgnoreCase("-tcp")) {
				tcp = true;
			} else if (args[i].equalsIgnoreCase("-r")) {
				i++;
				if (i >= args.length)
					usage();
				repeat = Integer.valueOf(args[i]);
			} else if (args[i].equalsIgnoreCase("-t")) {
				i++;
				if (i >= args.length)
					usage();
				duration = Integer.valueOf(args[i]);
			} else if (args[i].equalsIgnoreCase("-o")) {
				i++;
				if (i >= args.length)
					usage();
				filename = args[i];
			} else {
				System.err.println("Unknown option: " + args[i]);
				usage();
			}
		}

		if (!up && !down) {
			System.err.println("You have to specify at least one of -up or -down with true.");
			usage();
		}
		if (repeat <= 0) {
			System.err.println("-r must be at least 1.");
			usage();
		}

		BitTorrentWorker btw = new BitTorrentWorker(args[1], Integer.valueOf(args[2]));
		btw.confExp(up, down, tcp, port2, repeat, duration);
		btw.setFileOutput(filename);
		btw.start();
	}

}
