<?php

# Where to store logs and traces, use an absolute path!
$store_directory = "logs";

# The name of the local network interface
$eth = "eth0";

# Zip traces? (1 means yes)
$zip = 1;

##############################################################################
# Do not change anything below this line unless you know what you are doing! #
##############################################################################

# Global variables
$title = "Glasnost: Test if your ISP is manipulating BitTorrent traffic";
$server = $_SERVER['SERVER_ADDR'];
$version = "1.1";

# Subdir this script runs from
$subdir = substr($_SERVER['PHP_SELF'], 0, -6);

###############################################
# Process test results and display a summary
###############################################
function get_results() {
  global $title, $server, $store_directory, $subdir, $eth;

  # Check if results from the server-side are ready 
  # Failed, something went wrong
  if((! file_exists("$store_directory/latest.bt")) || (filesize("$store_directory/latest.bt") < 100)){

    unlink("$store_directory/running.bt");
    $id = getenv('REMOTE_ADDR');

    # Stats about user-satisfaction
    if(1){
      $name = gethostbyaddr($id);
      $fp = fopen("$store_directory/bt.stats", "a");
      fwrite($fp, time().";$id;$name;failed;no latest.bt\n");
      fclose($fp);
    }

    # Divert to error page
    header("Location: http://${server}/$subdir/error.php?id=${id}&msg=output%20not%20present&mid=12");

    return;
  }
  
  # Timestamp
  $ts = time();

  # Stats about user-satisfaction
  if(1){
    $id = getenv('REMOTE_ADDR');
    $name = gethostbyaddr($id);
    $fp = fopen("$store_directory/bt.stats", "a");
    fwrite($fp, "$ts;$id;$name;succeeded;\n");
    fclose($fp);
  }

  $tcp = 0;

  $sbtupc = 0;
  $sbtdownc = 0;
  $stcpupc = 0;
  $stcpdownc = 0;

  $thistest = -1;
  $sparam = "";

  $hostname = "";

  # Read and parse server-side log
  ###################################
  $handle = fopen("$store_directory/latest.bt", "r");
  while($tmp = fgets($handle)){
    $line = trim($tmp);
    @list ($trig, $data, $dir, $void, $dur, $port, $speed, $speed2, $void2) = split(" ", $line, 9);

    if(($hostname == "") && ($trig == "Client")){
      $hostname = $data;
      $ip = $dir;
    }
    elseif($trig == "Running"){
      if($thistest == 1){ $sbtupc++; }
      elseif($thistest == 2){ $sbtdownc++; }
      elseif($thistest == 3){ $stcpupc++; }
      elseif($thistest == 4){ $stcpdownc++; }

      if($data == "BT" && $dir == "upstream"){
	$thistest = 1;
	$sbtupp[$sbtupc] = $port;
      }
      elseif($data == "BT" && $dir == "downstream"){
	$thistest = 2;
	$sbtdownp[$sbtdownc] = $port;
      }
      elseif($data == "TCP" && $dir == "upstream"){
	$thistest = 3;
	$stcpupp[$stcpupc] = $port;
	$tcp = 1;
      }
      elseif($data == "TCP" && $dir == "downstream"){
	$thistest = 4;
	$stcpdownp[$stcpdownc] = $port;
	$tcp = 1;
      }
    }
    elseif($trig == "Transfered"){
      if($thistest == 1){
	$sbtup[$sbtupc] = $speed;
	$sparam = $sparam."sbtup${sbtupc}=$speed&";
      }
      elseif($thistest == 2){
	$sbtdown[$sbtdownc] = $speed;
	$sparam = $sparam."sbtdown${sbtdownc}=$speed&";
      }
      elseif($thistest == 3){
	$stcpup[$stcpupc] = $speed;
	$sparam = $sparam."stcpup${stcpupc}=$speed&";
      }
      elseif($thistest == 4){
	$stcpdown[$stcpdownc] = $speed;
	$sparam = $sparam."stcpdown${stcpdownc}=$speed&";
      }
    }
    elseif($trig == "FAILED:"){
      if($thistest == 1){
	$sbtup[$sbtupc] = $speed2;
	$sparam = $sparam."sbtup${sbtupc}=$speed2&";
      }
      elseif($thistest == 2){
	$sbtdown[$sbtdownc] = $speed2;
	$sparam = $sparam."sbtdown${sbtdownc}=$speed2&";
      }
      elseif($thistest == 3){
	$stcpup[$stcpupc] = $speed2;
	$sparam = $sparam."stcpup${stcpupc}=$speed2&";
      }
      elseif($thistest == 4){
	$stcpdown[$stcpdownc] = $speed2;
	$sparam = $sparam."stcpdown${stcpdownc}=$speed2&";
      }
    }
    elseif($trig == "Resets:"){
      if($thistest == 1){
	$sbtupr[$sbtupc] = $data; # Resets sent by the peer
	$sbtuprs[$sbtupc] = $port; # Resets sent by the server
	$sparam = $sparam."srbtup${sbtupc}=$data&srbtups${sbtupc}=$port&";
      }
      elseif($thistest == 2){
	$sbtdownr[$sbtdownc] = $data;
	$sbtdownrs[$sbtdownc] = $port;
	$sparam = $sparam."srbtdown${sbtdownc}=$data&srbtdowns${sbtdownc}=$port&";
      }
      elseif($thistest == 3){
	$stcpupr[$stcpupc] = $data;
	$stcpuprs[$stcpupc] = $port;
	$sparam = $sparam."srtcpup${stcpupc}=$data&srtcpups${stcpupc}=$port&";
      }
      elseif($thistest == 4){
	$stcpdownr[$stcpdownc] = $data;
	$stcpdownrs[$stcpdownc] = $port;
	$sparam = $sparam."srtcpdown${stcpdownc}=$data&srtcpdowns${stcpdownc}=$port&";
      }
    }
  }

  if($thistest == 1){ $sbtupc++; }
  elseif($thistest == 2){ $sbtdownc++; }
  elseif($thistest == 3){ $stcpupc++; }
  elseif($thistest == 4){ $stcpdownc++; }

  fclose($handle);

  # Save the log file from the server process
  $newname = sprintf("bt_%s_%s_%d.out", $ip, $hostname, $ts);
  system("mv $store_directory/latest.bt $store_directory/$newname >/dev/null 2>/dev/null &");
  #unlink("$store_directory/latest.bt"); 

  # Read in the results from the client side
  $param = "";
  while($p = each($_GET)){
    $param = $param."$p[0]=$p[1]&";
  }
  reset($_GET);

  # Store all results in a table for later use
  $fp = fopen("$store_directory/info.bt", "a");
  fwrite($fp, "$ts;$ip;$hostname;$param;$sparam;".$_GET['sysinfo']."\n");
  fclose($fp);


  # Now read what the client found
  ##################################
  $btupc = 0;
  $btdownc = 0;
  $tcpupc = 0;
  $tcpdownc = 0;

  # Three cases:
  # - Client returns a bps (can be 0 if nothing was transmitted)
  # - Client returns a bps with appended 'r': Something was transfered but connection was terminated by a RST
  # - Client returns 'reset': Nothing was transmitted, but the connection was terminated by a RST

  while(isset($_GET["btup${btupc}"])){
    if($_GET["btup${btupc}"] == "reset"){
      $btupr[$btupc] = 1;
    }
    else{
      $btup[$btupc] = $_GET["btup${btupc}"];
      if(preg_match('/\d*r/', $btup[$btupc])){
	$btup[$btupc] = substr($btup[$btupc], 0, -1); # Remove last character
      }
    }
    $btupc ++;
  }
  while(isset($_GET["btdown${btdownc}"])){
    if($_GET["btdown${btdownc}"] == "reset"){
      $btdownr[$btdownc] = 1;
    }
    else{
      $btdown[$btdownc] = $_GET["btdown${btdownc}"];
      if(preg_match('/\d*r/', $btdown[$btdownc])){
	$btdown[$btdownc] = substr($btdown[$btdownc], 0, -1); # Remove last character
      }
    }
    $btdownc ++;
  }
  while(isset($_GET["tcpup${tcpupc}"])){
    $tcp = 1;
    if($_GET["tcpup${tcpupc}"] == "reset"){
      $tcpupr[$tcpupc] = 1;
    }
    else{
      $tcpup[$tcpupc] = $_GET["tcpup${tcpupc}"];
      if(preg_match('/\d*r/', $tcpup[$tcpupc])){
	$tcpup[$tcpupc] = substr($tcpup[$tcpupc], 0, -1); # Remove last character
      }
    }
    $tcpupc ++;
  }
  while(isset($_GET["tcpdown${tcpdownc}"])){
    $tcp = 1;
    if($_GET["tcpdown${tcpdownc}"] == "reset"){
      $tcpdownr[$tcpdownc] = 1;
    }
    else{
      $tcpdown[$tcpdownc] = $_GET["tcpdown${tcpdownc}"];
      if(preg_match('/\d*r/', $tcpdown[$tcpdownc])){
	$tcpdown[$tcpdownc] = substr($tcpdown[$tcpdownc], 0, -1); # Remove last character
      }
    }
    $tcpdownc ++;
  }

  # The BT port used
  if(isset($_GET['btport'])){
    $btport = $_GET['btport'];
  }
  else{
    $btport = "unknown";
  }
  # We also ran on a non-standard port
  if(isset($_GET['port2'])){
    $port2 = $_GET['port2'];  
  }

  # The number of repeats for each experiment
  if(isset($_GET['repeat'])){
    $repeat = $_GET['repeat'];
  }
  else{
    if(isset($port2)){
      $repeat = (int)($btupc/2);
    }
    else{
      $repeat = $btupc;
    }
  }

# TMP store what we display
  ob_start();

  echo "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">
<html>
  <head>
    <title>$title</title>
    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">
    <link rel=\"stylesheet\" href=\"mpi.css\" type=\"text/css\">
    <link rel=\"shortcut icon\" href=\"./favicon.ico\" type=\"image/vnd.microsoft.icon\">
    <link rel=\"icon\" href=\"./favicon.ico\" type=\"image/vnd.microsoft.icon\"> 
  </head>

  <body>
    <a href=\"http://www.mpi-sws.mpg.de\"><img src=\"pics/mpisws-logo.png\" alt=\"MPI-SWS\" style=\"border:0px;background-color:white;vertical-align:bottom;\"></a>
    <h1 style=\"font-family:serif;\">$title</h1>

  <div style=\"width:75%;border:1px solid #69acff;padding:15px;\">
    
    <div style=\"border-bottom:1px solid #69acff;font-family:serif;font-weight:bold;\">
      Results for your host <i>($hostname - $ip)</i>:</div>    

";

  # Display results
  ####################

  echo "<h3 class=\"indent\" style=\"color:#777777;\">Is BitTorrent traffic on a well-known BitTorrent port ($btport) throttled?</h3>";

  # Client as a BT seeder on BT port
  $reset_count = 0;
  if($btdownc == 0){ $reset_count = -1;}
  else{
    for($i=0; $i<$repeat; $i++){
      # If both sides have seen resets (and the server at least 2)
      if(isset($sbtdownr[$i]) && isset($btupr[$i]) && ($sbtdownr[$i] >= 2) && ($sbtdownrs[$i] == 0)){ $reset_count++; }
    }
  }

  if($reset_count > 0){
    echo "<p class=\"indent\">
          <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
          <span style=\"color:red;font-weight:bold;\">$reset_count out of $repeat BitTorrent transfers were
          interrupted while uploading (seeding) using forged TCP RST packets.</span> 
          It seems like your ISP hinders you from uploading BitTorrent traffic to our test server.
          </p>
         ";
  }
  elseif($reset_count == 0){
    echo "<p class=\"indent\">
          <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
          <span style=\"font-weight:bold;\">The BitTorrent upload (seeding) worked.</span>
          Our tool was successful in uploading data using the BitTorrent protocol.
          </p>
          ";
  }


  $dtcpup = ""; $dbtup = "";
  $min_tcpup = -1;
  $max_btup = 0;

  # Compare BT and TCP rates
  if($tcp == 1){

    # Compare BT and TCP upload on BT port, only use measurements from the receiver!
    if($tcpupc > 0){
      for($i=0; $i<$repeat; $i++){
	if(isset($tcpup[$i]) && isset($stcpdown[$i])){
	  
	  if($tcpup[$i] > 0 && $stcpdown[$i] > 0){
	    if(($min_tcpup == -1) || ($min_tcpup > ($stcpdown[$i]/1000.0))){ $min_tcpup = (int)($stcpdown[$i]/1000.0); }
	    $dtcpup = sprintf("%s;%d", $dtcpup, (int)$stcpdown[$i]);
	  }
	  else{
	    $dtcpup = sprintf("%s;%d", $dtcpup, 0);
	  }
	}

	if(isset($btup[$i]) && isset($sbtdown[$i])){
	  if($btup[$i] > 0 && $sbtdown[$i] > 0){
	    if($max_btup < ($sbtdown[$i]/1000.0)) { $max_btup = (int)($sbtdown[$i]/1000.0); }
	    $dbtup = sprintf("%s;%d", $dbtup, (int)$sbtdown[$i]);
	  }
	  else{
	    $dbtup = sprintf("%s;%d", $dbtup, 0);
	  }
	}
      }

      if(($dtcpup != "") && ($dbtup != "")){
	$dtcpup = substr($dtcpup, 1); # Remove preceeding semicolon
	$dbtup = substr($dbtup, 1); # Remove preceeding semicolon
	
        # Now check if min/max is really bad and offer details
	if(($max_btup == 0) || ($min_tcpup == -1) || ($min_tcpup == 0)){
          # No valid data, say nothing (should we tell that there's a problem?)
	}
	elseif(($max_btup / $min_tcpup) < 0.5){
	  echo "<p class=\"indent2\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:red;font-weight:bold;\">Your ISP possibly rate limits your BitTorrent uploads.</span>
                  In our tests a TCP uploads achieved minimal $min_tcpup Kbps while a BitTorrent upload achieved
                  maximal $max_btup Kbps. You can find details
                  <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	}
	else{
	  echo "<p class=\"indent2\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:green;font-weight:bold;\">There's no indication that your ISP rate limits your BitTorrent uploads.</span>
                  In our tests a TCP upload achieved minimal $min_tcpup Kbps while a BitTorrent upload achieved
                  maximal $max_btup Kbps. You can find details <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	}
      }
    }
  }

  # Client as a BT downloader on BT port
  $reset_count = 0;
  if($btupc == 0){ $reset_count = -1; }
  else{
    for($i=0; $i<$repeat; $i++){
      # If both sides have seen resets (and the server at least 2)
      if(isset($sbtupr[$i]) && isset($btdownr[$i]) && ($sbtupr[$i] >= 2) && ($sbtuprs[$i] == 0)){ $reset_count++;}
    }
  }

  if($reset_count > 0){
    echo "<p class=\"indent\">
          <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
          <span style=\"color:red;font-weight:bold;\">$reset_count of $repeat BitTorrent transfers were 
          interrupted while downloading.</span>
          It seems like your ISP hinders you from downloading BitTorrent traffic from our test server.
          </p>
         ";
  }
  elseif($reset_count == 0){
    echo "<p class=\"indent\">
          <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
          <span style=\"font-weight:bold;\">The BitTorrent download worked.</span>
          Our tool was successful in downloading data using the BitTorrent protocol.
          </p>
          ";
  }

  $dtcpdown = ""; $dbtdown = "";
  $min_tcpdown = -1;
  $max_btdown = 0;

  # Compare BT and TCP rates
  if($tcp == 1){

    # Compare BT and TCP download on BT port (only use receiver measurements!)
    if($tcpdownc > 0){
      for($i=0; $i<$repeat; $i++){
	if(isset($tcpdown[$i]) && isset($stcpup[$i])){

	  if($tcpdown[$i] > 0 && $stcpup[$i] > 0){
	    if(($min_tcpdown == -1) || ($min_tcpdown > ($tcpdown[$i]/1000.0))){ $min_tcpdown = (int)($tcpdown[$i]/1000.0); }
	    $dtcpdown = sprintf("%s;%d", $dtcpdown, (int)$tcpdown[$i]);
	  }
	  else{
	    $dtcpdown = sprintf("%s;%d", $dtcpdown, 0);
	  }
	}

	if(isset($btdown[$i]) && isset($sbtup[$i])){

	  if($btdown[$i] > 0 && $sbtup[$i] > 0){
	    if($max_btdown < ($btdown[$i]/1000.0)) { $max_btdown = (int)($btdown[$i]/1000.0); }
	    $dbtdown = sprintf("%s;%d", $dbtdown, (int)$btdown[$i]);
	  }
	  else{
	    $dbtdown = sprintf("%s;%d", $dbtdown, 0);
	  }
	}
      }

      if(($dtcpdown != "") && ($dbtdown != "")){
	$dtcpdown = substr($dtcpdown, 1); # Remove preceeding semicolon
	$dbtdown = substr($dbtdown, 1); # Remove preceeding semicolon

        # Now check if min/max is really bad and offer details
	if(($max_btdown == 0) || ($min_tcpdown == -1) || ($min_tcpdown == 0)){
          # No valid data, say nothing (should we tell that there's a problem?)
	}
	elseif(($max_btdown / $min_tcpdown) < 0.5){
	  echo "<p class=\"indent2\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:red;font-weight:bold;\">Your ISP possibly rate limits your BitTorrent downloads.</span>
                  In our tests a TCP download achieved minimal ${min_tcpdown} Kbps while a BitTorrent download achieved
                  maximal $max_btdown Kbps. You can find details
                  <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	}
	else{
	  echo "<p class=\"indent2\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:green;font-weight:bold;\">There's no indication that your ISP rate limits your BitTorrent downloads.</span>
                  In our tests a TCP download achieved minimal $min_tcpdown Kbps while a BitTorrent download achieved
                  maximal $max_btdown Kbps. You can find details <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	}
      }
    }
  }
    


  # Client as a BT seeder on non-BT port
  if(isset($port2)){

    echo "<p>&nbsp;</p><h3 class=\"indent\" style=\"color:#777777;\">Is BitTorrent traffic on a non-standard BitTorrent port ($port2) throttled?</h3>";

    $reset_count = 0;
    if($btdownc == 0){ $reset_count = -1;}
    else{
      for($i=$repeat; $i<$btupc; $i++){
      # If both sides have seen resets (and the server at least 2)
	if(isset($sbtdownr[$i]) && isset($btupr[$i]) && ($sbtdownr[$i] >= 2) && ($sbtdownrs[$i] == 0)){ $reset_count++; }
      }
    }

    if($reset_count > 0){
      echo "<p class=\"indent\">
          <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
          <span style=\"color:red;font-weight:bold;\">$reset_count out of $repeat BitTorrent transfers were
          interrupted while uploading (seeding) using forged TCP RST packets.</span> 
          It seems like your ISP hinders you from uploading BitTorrent traffic to our test server.
          </p>
         ";
    }
    elseif($reset_count == 0){
      echo "<p class=\"indent\">
          <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
          <span style=\"font-weight:bold;\">The BitTorrent upload (seeding) worked.</span>
          Our tool was successful in uploading data using the BitTorrent protocol.
          </p>
          ";
    }
  
    $dtcpup2 = ""; $dbtup2 = "";
    $min_tcpup2 = -1;
    $max_btup2 = 0;

    if($tcp == 1){
      # Compare BT and TCP upload on non-BT port
      if($tcpupc > 0){
	for($i=$repeat; $i<$tcpupc; $i++){
	  if(isset($tcpup[$i]) && isset($stcpdown[$i])){
	  
	    if($tcpup[$i] > 0 && $stcpdown[$i] > 0){
	      if(($min_tcpup2 == -1) || ($min_tcpup2 > ($stcpdown[$i]/1000.0))){ $min_tcpup2 = (int)($stcpdown[$i]/1000.0); }
	      $dtcpup2 = sprintf("%s;%d", $dtcpup2, (int)$stcpdown[$i]);
	    }
	    else{
	      $dtcpup2 = sprintf("%s;%d", $dtcpup2, 0);
	    }
	  }

	  if(isset($btup[$i]) && isset($sbtdown[$i])){

	    if($btup[$i] > 0 && $sbtdown[$i] > 0){
	      if($max_btup2 < ($sbtdown[$i]/1000.0)) { $max_btup2 = (int)($sbtdown[$i]/1000.0); }
	      $dbtup2 = sprintf("%s;%d", $dbtup2, (int)$sbtdown[$i]);
	    }
	    else{
	      $dbtup2 = sprintf("%s;%d", $dbtup2, 0);
	    }
	  }
	}

	if(($dtcpup2 != "") && ($dbtup2 != "")){
	  $dtcpup2 = substr($dtcpup2, 1); # Remove preceeding semicolon
	  $dbtup2 = substr($dbtup2, 1); # Remove preceeding semicolon

          # Now check if min/max is really bad and offer details
	  if(($max_btup2 == 0) || ($min_tcpup2 == -1) || ($min_tcpup2 == 0)){
            # No valid data, say nothing (should we tell that there's a problem?)
	  }
	  elseif(($max_btup2 / $min_tcpup2) < 0.5){
	    echo "<p class=\"indent2\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:red;font-weight:bold;\">Your ISP possibly rate limits your BitTorrent uploads.</span>
                  In our tests a TCP uploads achieved minimal $min_tcpup2 Kbps while a BitTorrent upload achieved
                  maximal $max_btup2 Kbps. You can find details
                  <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	  }
	  else{
	    echo "<p class=\"indent2\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:green;font-weight:bold;\">There's no indication that your ISP rate limits your BitTorrent uploads.</span>
                  In our tests a TCP download achieved minimal $min_tcpup2 Kbps while a BitTorrent download achieved
                  maximal $max_btup2 Kbps. You can find details <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	  }
	}
      }
    }

    # Client as a BT downloader on non-BT port
    $reset_count = 0;
    if($btupc == 0){ $reset_count = -1; }
    else{
      for($i=$repeat; $i<$btdownc; $i++){
        # If both sides have seen resets (and the server at least 2)
	if(isset($sbtupr[$i]) && isset($btdownr[$i]) && ($sbtupr[$i] >= 2) && ($sbtuprs[$i] == 0)){ $reset_count++;}
      }
    }

    if($reset_count > 0){
      echo "<p class=\"indent\">
          <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
          <span style=\"color:red;font-weight:bold;\">$reset_count of $repeat BitTorrent transfers were 
          interrupted while downloading.</span>
          It seems like your ISP hinders you from downloading BitTorrent traffic from our test server.
          </p>
         ";
    }
    elseif($reset_count == 0){
      echo "<p class=\"indent\">
          <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
          <span style=\"font-weight:bold;\">The BitTorrent download worked.</span>
          Our tool was successful in downloading data using the BitTorrent protocol.
          </p>
          ";
    }

    $dtcpdown2 = ""; $dbtdown2 = "";
    $min_tcpdown2 = -1;
    $max_btdown2 = 0;

    if($tcp == 1){

      # Compare BT and TCP download on non-BT port
      if($tcpdownc > 0){
	for($i=$repeat; $i<$tcpdownc; $i++){
	  if(isset($tcpdown[$i]) && isset($stcpup[$i])){
	  
	    if($tcpdown[$i] > 0 && $stcpup[$i]){
	      if(($min_tcpdown2 == -1) || ($min_tcpdown2 > ($tcpdown[$i]/1000.0))){ $min_tcpdown2 = (int)($tcpdown[$i]/1000.0); }
	      $dtcpdown2 = sprintf("%s;%d", $dtcpdown2, (int)$tcpdown[$i]);
	    }
	    else{
	      $dtcpdown2 = sprintf("%s;%d", $dtcpdown2, 0);
	    }
	  }

	  if(isset($btdown[$i]) && isset($sbtup[$i])){

	    if($btdown[$i] > 0 && $sbtup[$i] >  0){
	      if($max_btdown2 < ($btdown[$i]/1000.0)) { $max_btdown2 = (int)($btdown[$i]/1000.0); }
	      $dbtdown2 = sprintf("%s;%d", $dbtdown2, (int)$btdown[$i]);
	    }
	    else{
	      $dbtdown2 = sprintf("%s;%d", $dbtdown2, 0);
	    }
	  }
	}

	if(($dtcpdown2 != "") && ($dbtdown2 != "")){
	  $dtcpdown2 = substr($dtcpdown2, 1); # Remove preceeding semicolon
	  $dbtdown2 = substr($dbtdown2, 1); # Remove preceeding semicolon

          # Now check if min/max is really bad and offer details
	  if(($max_btdown2 == 0) || ($min_tcpdown2 == -1) || ($min_tcpdown2 == 0)){
	    # No valid data, say nothing (should we tell that there's a problem?)
	  }
	  elseif(($max_btdown2 / $min_tcpdown2) < 0.5){
	    echo "<p class=\"indent2\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:red;font-weight:bold;\">Your ISP possibly rate limits your BitTorrent downloads.</span>
                  In our tests a TCP download achieved minimal $min_tcpdown2 Kbps while a BitTorrent download achieved
                  maximal $max_btdown2 Kbps. You can find details
                  <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	  }
	  else{
	    echo "<p class=\"indent2\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:green;font-weight:bold;\">There's no indication that your ISP rate limits your BitTorrent downloads.</span>
                  In our tests a TCP download achieved minimal $min_tcpdown2 Kbps while a BitTorrent download achieved
                  maximal $max_btdown2 Kbps. You can find details <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	  }
	}
      }
    }

    if((($dtcpdown != "") && ($dtcpdown2 != "")) || (($dtcpup != "") && ($dtcpup2 != ""))){
      echo "<p>&nbsp;</p>
            <h3 class=\"indent\" style=\"color:#777777;\">Is TCP traffic on a well-known BitTorrent port ($btport) throttled?</h3>";
           
      if(($dtcpdown != "") && ($dtcpdown2 != "")){

	if(($min_tcpdown == -1) || ($min_tcpdown2 == -1) || ($min_tcpdown2 == 0)){}
	elseif(($min_tcpdown / $min_tcpdown2) < 0.5){
	    echo "<p class=\"indent\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:red;font-weight:bold;\">Your ISP possibly rate limits all downloads at port $btport.</span>
                  In our test, a TCP download on a BitTorrent port achieved at least $min_tcpdown Kbps while a TCP download on a
                  non-BitTorrent port achieved at least $min_tcpdown2 Kbps.
                  You can find details
                  <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	  }
	  else{
	    echo "<p class=\"indent\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:green;font-weight:bold;\">There's no indication that your ISP rate limits all downloads at port $btport.</span>
                  In our test, a TCP download on a BitTorrent port achieved at least $min_tcpdown Kbps while a TCP download on a
                  non-BitTorrent port achieved at least $min_tcpdown2 Kbps.
                  You can find details <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	  }
      }
      if(($dtcpup != "") && ($dtcpup2 != "")){
	if(($min_tcpup == -1) || ($min_tcpup2 == -1) || ($min_tcpup2 == 0)){}
	elseif(($min_tcpup / $min_tcpup2) < 0.5){
	    echo "<p class=\"indent\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:red;font-weight:bold;\">Your ISP possibly rate limits all uploads at port $btport.</span>
                  In our test, a TCP upload on a BitTorrent port achieved at least $min_tcpup Kbps while a TCP upload on a
                  non-BitTorrent port achieved at least $min_tcpup2 Kbps.
                  You can find details
                  <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	  }
	  else{
	    echo "<p class=\"indent\">
                  <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
                  <span style=\"color:green;font-weight:bold;\">There's no indication that your ISP rate limits all uploads at port $btport.</span>
                  In our test, a TCP upload on a BitTorrent port achieved at least $min_tcpup Kbps while a TCP upload on a
                  non-BitTorrent port achieved at least $min_tcpup2 Kbps.
                  You can find details <a href=\"#\" onclick=\"openDetails()\">here</a>.
	          </p>
                 ";
	  }
      }
    }
  }

  echo "
    <p>&nbsp;</p>
    <p style=\"padding-top:5px;border-top:1px solid #69acff;\">
       For details on our research on broadband networks please refer to our
       <a href=\"http://broadband.mpi-sws.org/transparency/\">network transparency project webpage</a></p>

    <p>In case you have questions about this tool or our research, please contact us: 
<img src=\"http://broadband.mpi-sws.org/pics/email.png\" alt=\"broadband @at@ mpi-sws mpg de\" style=\"margin-left:5px;margin-bottom:-6px;\"></p>
   </div>
";

  $params = "";
  if(isset($btport)){ $params = $params."btport=${btport}&"; }
  if(isset($port2)){ $params = $params."port2=${port2}&"; }
  if($dtcpdown != ""){ $params = $params."tcpdown=${dtcpdown}&"; }
  if($dbtdown != ""){ $params = $params."btdown=${dbtdown}&"; }
  if($dtcpup != ""){ $params = $params."tcpup=${dtcpup}&"; }
  if($dbtup != ""){ $params = $params."btup=${dbtup}&"; }
  if($dtcpdown2 != ""){ $params = $params."tcpdown2=${dtcpdown2}&"; }
  if($dbtdown2 != ""){ $params = $params."btdown2=${dbtdown2}&"; }
  if($dtcpup2 != ""){ $params = $params."tcpup2=${dtcpup2}&"; }
  if($dbtup2 != ""){ $params = $params."btup2=${dbtup2}&"; }

  if($params != ""){
    $params = $params."details=yes";

    echo"<script type=\"text/javascript\">
         <!--
          function openDetails(){
            newWin = window.open(\"http://${server}/$subdir/bt.php?${params}&\", \"_blank\", \"\");
            newWin.focus();
          }
         -->
        </script>
        ";
  }

  echo" </body></html>";

  # Store what we showed to the user
  $newname = sprintf("bt_%s_%s_%d.html", $ip, $hostname, $ts);
  $fp = fopen("$store_directory/$newname", "w");
  fwrite($fp, ob_get_contents());
  fclose($fp);


  # Now send all content to the user
  ob_end_flush();

  unlink("$store_directory/running.bt");

  exit(0);
}

###############################################################################
# For the bandwidth comparison, display detailed tables with individual results
###############################################################################
function show_details() {
  global $title;

  if(isset($_GET['btport'])){
    $btport = $_GET['btport'];
  }
  if(isset($_GET['port2'])){
    $port2 = $_GET['port2'];
  }

  # BT port
  if(isset($_GET['tcpdown'])){
    $tcpdown = split(';', $_GET['tcpdown']);
  }
  if(isset($_GET['tcpup'])){
    $tcpup = split(';', $_GET['tcpup']);
  }
  if(isset($_GET['btdown'])){
    $btdown = split(';', $_GET['btdown']);
  }
  if(isset($_GET['btup'])){
    $btup = split(';', $_GET['btup']);
  }

  # Non-BT port
  if(isset($_GET['tcpdown2'])){
    $tcpdown2 = split(';', $_GET['tcpdown2']);
  }
  if(isset($_GET['tcpup2'])){
    $tcpup2 = split(';', $_GET['tcpup2']);
  }
  if(isset($_GET['btdown2'])){
    $btdown2 = split(';', $_GET['btdown2']);
  }
  if(isset($_GET['btup2'])){
    $btup2 = split(';', $_GET['btup2']);
  }

  echo "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">
<html>
  <head>
    <title>$title</title>
    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">
    <link rel=\"stylesheet\" href=\"mpi.css\" type=\"text/css\">
    <link rel=\"shortcut icon\" href=\"./favicon.ico\" type=\"image/vnd.microsoft.icon\">
    <link rel=\"icon\" href=\"./favicon.ico\" type=\"image/vnd.microsoft.icon\"> 
  </head>

  <body>
    <a href=\"http://www.mpi-sws.mpg.de\"><img src=\"pics/mpisws-logo.png\" alt=\"MPI-SWS\" style=\"border:0px;background-color:white;vertical-align:bottom;\"></a>
    <h1 style=\"font-family:serif;\">$title</h1>

  <div style=\"width:75%;border:1px solid #69acff;padding:15px;\">
    
    <div style=\"border-bottom:1px solid #69acff;font-family:serif;font-weight:bold;\">
      Detailed test results:</div>    
";

  if(isset($btport)){

    $firstentry = 1;
    $i = 0;
    while(isset($tcpdown[$i]) && isset($btdown[$i]) && $tcpdown[$i] != "" && $btdown[$i] != ""){

      if($tcpdown[$i] > 0 && $btdown[$i] > 0){ # A value of 0 indicates a measurement failure

	if($firstentry == 1){
	  echo "<h3 class=\"indent\" style=\"color:#777777;\">Is BitTorrent traffic on a well-known BitTorrent port ($btport) throttled?</h3>";

	  echo "<p class=\"indent\">
           <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
           Completed BitTorrent and TCP transfers using the BitTorrent port $btport:</p>
           <p style=\"padding-left:30px;\">
           <table style=\"border:0px;background-color:black;\" cellspacing=\"1\">
            <tr>
             <th class=\"box\">Transfer</th>
             <th class=\"box\">Speed TCP</th>
             <th class=\"box\">Speed BitTorrent</th>
             <th class=\"box\">Conclusion</th>
            </tr>
          ";
	  $firstentry = 0;
	}

	printf("<tr><td class=\"box\">Download #$i</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
              ", (int)$tcpdown[$i]/1000, (int)$btdown[$i]/1000);
	

	if((($btdown[$i]/$tcpdown[$i]) < 0.5)){
	  echo "<td class=\"box\" style=\"color:red;font-weight:bold;\">Potential rate limiting</td>";
	}
	else{
	  echo "<td class=\"box\" style=\"color:green;font-weight:bold;\">No rate limiting</td>";
	}
	echo "</tr>";
      }
      $i++;
    }
    
    $i = 0;
    while(isset($tcpup[$i]) && isset($btup[$i]) && $tcpup[$i] != "" && $btup[$i] != ""){

      if($tcpup[$i] > 0 && $btup[$i] > 0){ # A value of 0 indicates a measurement failure

	if($firstentry == 1){
	  echo "<h3 class=\"indent\" style=\"color:#777777;\">Is BitTorrent traffic on a well-known BitTorrent port ($btport) throttled?</h3>";

	  echo "<p class=\"indent\">
           <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
           Completed BitTorrent and TCP transfers using the BitTorrent port $btport:</p>
           <p style=\"padding-left:30px;\">
           <table style=\"border:0px;background-color:black;\" cellspacing=\"1\">
            <tr>
             <th class=\"box\">Transfer</th>
             <th class=\"box\">Speed TCP</th>
             <th class=\"box\">Speed BitTorrent</th>
             <th class=\"box\">Conclusion</th>
            </tr>
          ";
	  $firstentry = 0;
	}

	printf("<tr><td class=\"box\">Upload #$i</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
              ", (int)$tcpup[$i]/1000, (int)$btup[$i]/1000);
	

	if((($btup[$i]/$tcpup[$i]) < 0.5)){
	  echo "<td class=\"box\" style=\"color:red;font-weight:bold;\">Potential rate limiting</td>";
	}
	else{
	  echo "<td class=\"box\" style=\"color:green;font-weight:bold;\">No rate limiting</td>";
	}
	echo "</tr>";
      }
      $i++;
    }

    if($firstentry == 0){
      echo "</table>";
      echo "<p>&nbsp;</p>";
    }
  }

  if(isset($port2)){

    $firstentry = 1;
    $i = 0;
    while(isset($tcpdown2[$i]) && isset($btdown2[$i]) && $tcpdown2[$i] != "" && $btdown2[$i] != ""){

      if($tcpdown2[$i] > 0 && $btdown2[$i] > 0){ # A value of 0 indicates a measurement failure

	if($firstentry == 1){
	  echo "<h3 class=\"indent\" style=\"color:#777777;\">Is BitTorrent traffic on a non-standard BitTorrent port ($port2) throttled?</h3>";
	  echo "<p class=\"indent\">
           <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
           Completed BitTorrent and TCP transfers using the non-BitTorrent port $port2:</p>
           <p style=\"padding-left:30px;\">
           <table style=\"border:0px;background-color:black;\" cellspacing=\"1\">
            <tr>
             <th class=\"box\">Transfer</th>
             <th class=\"box\">Speed TCP</th>
             <th class=\"box\">Speed BitTorrent</th>
             <th class=\"box\">Conclusion</th>
            </tr>
          ";
	  $firstentry = 0;
	}

	printf("<tr><td class=\"box\">Download #$i</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
              ", (int)$tcpdown2[$i]/1000, (int)$btdown2[$i]/1000);
	

	if((($btdown2[$i]/$tcpdown2[$i]) < 0.5)){
	  echo "<td class=\"box\" style=\"color:red;font-weight:bold;\">Potential rate limiting</td>";
	}
	else{
	  echo "<td class=\"box\" style=\"color:green;font-weight:bold;\">No rate limiting</td>";
	}
	echo "</tr>";
      }
      $i++;
    }
    
    $i = 0;
    while(isset($tcpup2[$i]) && isset($btup2[$i]) && $tcpup2[$i] != "" && $btup2[$i] != ""){


      if($tcpup2[$i] > 0 && $btup2[$i] > 0){ # A value of 0 indicates a measurement failure

	if($firstentry == 1){
	  echo "<h3 class=\"indent\" style=\"color:#777777;\">Is BitTorrent traffic on a non-standard BitTorrent port ($port2) throttled?</h3>";
	  echo "<p class=\"indent\">
           <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
           Completed BitTorrent and TCP transfers using the non-BitTorrent port $port2:</p>
           <p style=\"padding-left:30px;\">
           <table style=\"border:0px;background-color:black;\" cellspacing=\"1\">
            <tr>
             <th class=\"box\">Transfer</th>
             <th class=\"box\">Speed TCP</th>
             <th class=\"box\">Speed BitTorrent</th>
             <th class=\"box\">Conclusion</th>
            </tr>
          ";
	  $firstentry = 0;
	}

	printf("<tr><td class=\"box\">Upload #$i</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
              ", (int)$tcpup2[$i]/1000, (int)$btup2[$i]/1000);
	

	if((($btup2[$i]/$tcpup2[$i]) < 0.5)){
	  echo "<td class=\"box\" style=\"color:red;font-weight:bold;\">Potential rate limiting</td>";
	}
	else{
	  echo "<td class=\"box\" style=\"color:green;font-weight:bold;\">No rate limiting</td>";
	}
	echo "</tr>";
      }
      $i++;
    }

    if($firstentry == 0){
      echo "</table>";
      echo "<p>&nbsp;</p>";
    }


    # Now check if traffic is rate limited if it used a well-known BT port

    $firstentry = 1;
    $i = 0;
    while(isset($tcpdown2[$i]) && isset($tcpdown[$i]) && $tcpdown2[$i] != "" && $tcpdown[$i] != ""){

      if($tcpdown2[$i] > 0 && $tcpdown[$i] > 0){ # A value of 0 indicates a measurement failure

	if($firstentry == 1){
	  echo "<h3 class=\"indent\" style=\"color:#777777;\">Is TCP traffic on a well-known BitTorrent port ($btport) throttled?</h3>";
    
	  echo "<p class=\"indent\">
           <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
           Comparing TCP transfers using the well-know BitTorrent port $btport and the non-BitTorrent port $port2:</p>
           <p style=\"padding-left:30px;\">
           <table style=\"border:0px;background-color:black;\" cellspacing=\"1\">
            <tr>
             <th class=\"box\">TCP Transfer</th>
             <th class=\"box\">BitTorrent port</th>
             <th class=\"box\">Non-BitTorrent port</th>
             <th class=\"box\">Conclusion</th>
            </tr>
          ";
	  $firstentry = 0;
	}

	printf("<tr><td class=\"box\">Download #$i</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
              ", (int)$tcpdown[$i]/1000, (int)$tcpdown2[$i]/1000);
	

	if((($tcpdown[$i]/$tcpdown2[$i]) < 0.5)){
	  echo "<td class=\"box\" style=\"color:red;font-weight:bold;\">Potential rate limiting</td>";
	}
	else{
	  echo "<td class=\"box\" style=\"color:green;font-weight:bold;\">No rate limiting</td>";
	}
	echo "</tr>";
      }
      $i++;
    }

    $i = 0;
    while(isset($tcpup2[$i]) && isset($tcpup[$i]) && $tcpup2[$i] != "" && $tcpup[$i] != ""){

      if($tcpup2[$i] > 0 && $tcpup[$i] > 0){ # A value of 0 indicates a measurement failure

	if($firstentry == 1){
	  echo "<h3 class=\"indent\" style=\"color:#777777;\">Is TCP traffic on a well-known BitTorrent port ($btport) throttled?</h3>";

	  echo "<p class=\"indent\">
           <img src=\"pics/square.png\" alt=\"*\" style=\"margin-right:5px;margin-left:5px;margin-bottom:3px;\">
           Comparing TCP transfers using the well-know BitTorrent port $btport and the non-BitTorrent port $port2:</p>
           <p style=\"padding-left:30px;\">
           <table style=\"border:0px;background-color:black;\" cellspacing=\"1\">
            <tr>
             <th class=\"box\">TCP Transfer</th>
             <th class=\"box\">BitTorrent port</th>
             <th class=\"box\">Non-BitTorrent port</th>
             <th class=\"box\">Conclusion</th>
            </tr>
          ";
	  $firstentry = 0;
	}

	printf("<tr><td class=\"box\">Upload #$i</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
                  <td style=\"text-align:right;\" class=\"box\">%d Kbps</td>
              ", (int)$tcpup[$i]/1000, (int)$tcpup2[$i]/1000);
	

	if((($tcpup[$i]/$tcpup2[$i]) < 0.5)){
	  echo "<td class=\"box\" style=\"color:red;font-weight:bold;\">Potential rate limiting</td>";
	}
	else{
	  echo "<td class=\"box\" style=\"color:green;font-weight:bold;\">No rate limiting</td>";
	}
	echo "</tr>";
      }
      $i++;
    }

    if($firstentry == 0){
      echo "</table>";
      echo "<p>&nbsp;</p>";
    }
  }

  echo "
    <p><i>Note that it is hard to detect rate limiting and that there are many explanations why a transfer can observe 
       lower throughput than another transfer. We run the BitTorrent and TCP transfer back-to-back and at least two
       times to add more confidence to our measurements.</i>
    </p>

    <p>&nbsp;</p>
    <p style=\"padding-top:5px;border-top:1px solid #69acff;\">
       For details on our research on broadband networks please refer to our
       <a href=\"http://broadband.mpi-sws.org/transparency/\">network transparency project webpage</a></p>

    <p>In case you have questions about this tool or our research, please contact us: 
<img src=\"http://broadband.mpi-sws.org/pics/email.png\" alt=\"broadband @at@ mpi-sws mpg de\" style=\"margin-left:5px;margin-bottom:-6px;\"></p>
   </div>
  </body>
</html>
";

  exit(0);
}

############################################################################################
# Start the measurement: Fork the server-side infrastructure and load the client-side applet
############################################################################################
function start_measurement($mode) {
  global $title, $store_directory, $subdir, $eth, $zip;
  $id = getenv('REMOTE_ADDR'); # This might be a proxy!
#  $ip = $_GET['id'];

  # Get all parameters for the measurement
  $port = 6881;
  if(isset($_GET['port'])){
    $port = $_GET['port'];
  }

  $port = -1;
  if(isset($_GET['port']) && is_numeric($_GET['port'])){
    $port = $_GET['port'];
  }

  $port2 = -1;
  if(isset($_GET['port2'])){
    $port2 = $_GET['port2'];
  }

  $tcp = "false";
  if(isset($_GET['tcp']) && ($_GET['tcp'] == "yes")){
    $tcp = "true";
  }

  $num_repeat = 2;
  if(isset($_GET['repeat'])){
    $num_repeat = $_GET['repeat'];
  }

  $duration = -1;
  if(isset($_GET['duration'])){
    $duration = $_GET['duration'];
  }

  $down = "false";
  if(isset($_GET['down']) && ($_GET['down'] == "yes")){
    $down = "true";
  }

  $up = "true";
  if(isset($_GET['up']) && ($_GET['up'] == "yes")){
    $up = "true";
  }

  # Stats about user-satisfaction
  if(1){
    $name = gethostbyaddr($id);
    $fp = fopen("$store_directory/bt.stats", "a");
    fwrite($fp, time().";$id;$name;measuring\n");
    fclose($fp);
  }
  

   # Start server and wait for client to log on
#   exec("/var/www/$subdir/bt_client -h $id -i $eth -o $store_directory/bt.dump >$store_directory/latest.bt 2>stderr &");
  $docroot = $_SERVER['DOCUMENT_ROOT'];
#  exec("$docroot/$subdir/bt_client -i $eth -o $store_directory/bt.dump >$store_directory/latest.bt 2>>$store_directory/stderr &");
  if($zip == 1){
    exec("(cd $store_directory ; $docroot/$subdir/bt_client -i $eth -z) >$store_directory/latest.bt 2>>$store_directory/stderr &");
  }
  else{
    exec("(cd $store_directory ; $docroot/$subdir/bt_client -i $eth) >$store_directory/latest.bt 2>>$store_directory/stderr &");
  }
  
  # Mac gets a special jar (Mac Java needs privileges for select() calls)
  $jar = "BitTorrentClient.jar";
  if (eregi("mac", $_SERVER['HTTP_USER_AGENT'])){ $jar = "BitTorrentClientMac.jar"; }
        
   # Display HTML page and load Applet
   echo "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">
<html>
  <head>
    <title>$title</title>
    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">
    <meta http-equiv=\"cache-control\" content=\"no-cache\">
    <meta http-equiv=\"pragma\" content=\"no-cache\">
    <link rel=\"stylesheet\" href=\"mpi.css\" type=\"text/css\">
    <link rel=\"shortcut icon\" href=\"./favicon.ico\" type=\"image/vnd.microsoft.icon\">
    <link rel=\"icon\" href=\"./favicon.ico\" type=\"image/vnd.microsoft.icon\"> 
  </head>

  <body>
    <a href=\"http://www.mpi-sws.mpg.de\"><img src=\"pics/mpisws-logo.png\" alt=\"MPI-SWS\" style=\"border:0px;background-color:white;vertical-align:bottom;\"></a>
    <h1 style=\"font-family:serif;\">$title</h1>

    <table style=\"border:1px solid #69acff;padding:15px;\">
     <tr><td style=\"padding-bottom:20px;text-align:center;\">
        <b>Testing...</b>
     </td></tr><tr><td style=\"text-align:center;font-weight:bold;color:red;\">
       <applet archive=\"$jar\" code=\"BitTorrentClient\" height=\"40\" width=\"200\" alt=\"Your Java plugin is not working!\">
       <param name=\"ID\" value=\"$id\">
       <param name=\"port\" value=\"$port\">
       <param name=\"up\" value=\"$up\">
       <param name=\"down\" value=\"$down\">
       <param name=\"tcp\" value=\"$tcp\">
       <param name=\"port2\" value=\"$port2\">
       <param name=\"duration\" value=\"$duration\">
       <param name=\"repeat\" value=\"$num_repeat\">
       <param name=\"cache_option\" value=\"no\">
       <param name=\"cache_version\" value=\"2.0\">
       <param name=\"errorpage\" value=\"$subdir/error.php\">
       <param name=\"testpage\" value=\"$subdir/bt.php\">
       <b style=\"color:red\">Your Java plugin is not working!</b>
       </applet>
     </td></tr><tr><td style=\"padding-top:20px;text-align:center;\">
       <b>Please wait while a BitTorrent transfer is emulated.</b>
     </td></tr></table>

     <p style=\"width:400px;\"><i>
        You need a Java plugin installed to use this tool (you can get the Java plugin for free
        <a href=\"http://java.sun.com/javase/downloads/index.jsp\">here</a>).<br>
        If the Java applet above does not start and you do not see a blue progress bar appearing, 
        please check your Java plugin installation.<br>
        Otherwise you can contact 
        <a href=\"http://www.mpi-sws.org/~mdischin/\">Marcel Dischinger</a>.</i></p>
";

   echo "</div></body></html>";

   exit(0);
}

function redirect($new_param) {

  $params = "";
  if(isset($new_param) && ($new_param != "")){
    $params = "$new_param&";
  }
  # Read in all parameters
  while($p = each($_GET)){
    $params = $params."$p[0]=$p[1]&";
  }
  reset($_GET);

  header("Location: http://broadband.mpi-sws.mpg.de/transparency/bttest.php?$params");

  exit(0);
}



#############
# Main loop #
#############

# Check if directory to store logs and traces is present
if(!is_dir($store_directory)){
     print "Internal error: '$store_directory' does not exist or is not a directory.\n";
     print "Please create this directory.\n";
     exit(1);
}

if(isset($_GET['version'])){
   echo "$version\n";
   echo "$store_directory\n";
   exit(0);
}
# Client wants to be measured, check if we can do it
elseif(isset($_GET['measure']) && ($_GET['measure'] == "yes") && isset($_GET['ip'])) { 

  # Failure
  if(isset($_GET['expert']) && ($_GET['expert'] == "on")){
    if(! isset($_GET['up']) && ! isset($_GET['down'])){
      redirect(null);
    }
  }

  $ip = getenv('REMOTE_ADDR'); # This might also be a proxy address!

  # Calculate how long this measurement might take
  $exp_length = 0;
  $num = 1;
  
  # Note: The order is important here!
  if(isset($_GET['down']) && ($_GET['down'] == "yes")){
    $num++;
  }
  if(isset($_GET['tcp']) && ($_GET['tcp'] == "yes")){
    $exp_length = $num * 10; # This is for setting up a connection
    $num *= 2;
  }
  if(isset($_GET['port2']) && is_numeric($_GET['port2'])){
    $num *= 2;
  }
  if(isset($_GET['repeat'])){
    $num *= $_GET['repeat'];
  }
  else{
    $num *= 2; # default is to repeat each experiement once
  }
  if(isset($_GET['duration'])){
    $duration = $_GET['duration'];
  }
  else{
    $duration = 20;
  }
  # $duration seconds per experiment, 4 seconds waiting between experiments
  $exp_length += (($duration+4) * $num); # In seconds 

  # Check if someone else is already measured
  if(file_exists("$store_directory/running.bt")){
    
    $fp = fopen("$store_directory/running.bt", "r+");
    $line = fgets($fp);
    $line = trim($line);
    list ($timestamp, $id, $length, ) = split(" ", $line, 4);

    if((!isset($length)) || (!is_numeric($length))){
      $length = 90;
    }
    if((((time() - $timestamp) > $length) || (time() - $timestamp) > 90)){

      $out = `ps xea | grep bt_client | grep -v grep > /dev/null ; echo \$?`;
      $out = trim($out);

      # After 30min, bt_client just hangs => Sometimes the transfer is so slow that it takes forever (trans. proxy as in UMTS?)
      # Check when the last modification time of latest.bt was!
      if(($out == 0) && ((time() - $timestamp) > 30*60)){
	system("killall bt_client");
	$out = 1;
      }

      # No bt_client process is running and latest.bt is too small to contain a succeeded transfer
      if(((time() - $timestamp) > $length) || (($out != 0) && ((! file_exists("${store_directory}/latest.bt")) || (filesize("${store_directory}/latest.bt") < 200)))){

	# move log files for failure analysis
	$newname = sprintf("failed_%s_%s_%d", $id, $id, $timestamp); 

	if(file_exists("${store_directory}/latest.bt")){
	  system("mv ${store_directory}/latest.bt ${store_directory}/${newname}.out");
	}

	ftruncate($fp, 0);
	fwrite($fp, time()." ".$_GET['ip']." ".$exp_length." ".$ip);
	fclose($fp);

	start_measurement(0);
      }
      else{
	fclose($fp);
	redirect("busy=1");
      }
    }
    else{
      fclose($fp);
      redirect("busy=1");
    }
  }
  else{
    $fp = fopen("$store_directory/running.bt", "w");
    fwrite($fp, time()." ".$_GET['ip']." ".$exp_length." ".$ip);
    fclose($fp);

    start_measurement(0);
  }
}
# Measurement done, present results
else if(isset($_GET['done']) && $_GET['done'] == "yes" && isset($_GET['id']) && isset($_GET['btport'])){

  # Make sure the caller is really the last runner
  if(!file_exists("$store_directory/running.bt")){
    # Divert to error page
    $id = $_GET['id'];
    header("Location: http://${server}/$subdir/error.php?id=${id}&msg=not%20running&mid=14");
  }
  else{
    $fp = fopen("$store_directory/running.bt", "r");
    $line = fgets($fp);
    $line = trim($line);
    list ($timestamp, $id, $length, ) = split(" ", $line, 4);
    if($id != $_GET['id']){
      $id = $_GET['id'];
      header("Location: http://${server}/$subdir/error.php?id=${id}&msg=not%20running&mid=14");
      exit(1);
    }
  }

  get_results();
}
# Show details to a given result
else if($_GET['details'] == "yes" && (isset($_GET['btport']) || isset($_GET['port2']))){
   show_details();
}
# Display start page
else { 
  redirect(null);
} 
