#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>
#include <ctype.h>

#include <linux/if_packet.h>
#include <net/ethernet.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <fcntl.h>
#include <pcap.h>
#include <zlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>

#define MAX(a, b) (a > b ? a : b)
#define MIN(a, b) (a < b ? a : b)

#define SETUP 0
#define BT 1
#define TCP 2

#define UPSTREAM 0
#define DOWNSTREAM 1

#define VERSION "2.2.1"

#define panic(a...) do { fprintf(stderr, a); fprintf(stderr, "\n"); fprintf(stderr, "bt_client: panic!\n"); exit(1); } while (0)
#define log(b...) do { fprintf(stdout, b); fprintf(stdout, "\n"); fflush(stdout); } while (0)
#define log2(b) do { fprintf(stdout, "%lld ", (getTimeMicros()/1000LL)); fprintf(stdout, b); fprintf(stdout, "\n"); fflush(stdout); } while (0)
#define log2i(b,c) do { fprintf(stdout, "%lld ", (getTimeMicros()/1000LL)); fprintf(stdout, b); fprintf(stdout, "%d", (c)); fprintf(stdout, "\n"); fflush(stdout); } while (0)
#define log2x(b...) do { fprintf(stdout, "%lld ", (getTimeMicros()/1000LL)); fprintf(stdout, b); fprintf(stdout, "\n"); fflush(stdout); } while (0)
#define error(b...) do { fprintf(stdout, "ERROR: "); fprintf(stdout, b); fprintf(stdout, "\n"); fflush(stdout); } while (0)
//#define log(a, b...) do { } while (0)


pcap_dumper_t *myDumper = NULL;
pcap_t *myPcap = NULL;
int peer = -1;
int read_socket = -1; // Raw socket to sniff from
int sock; // TCP socket to accept connections to
int command_sock; // TCP socket to receive commands on
int command_port = 19980;

char dumpFile[500] = "";

const int pieceSize = 262144; // Piece size to send/receive data blocks

int isTerminated = 0;
int transferSuccessful = 1;
int resets = 0, ownResets = 0;

int token = 0; // Anonymize peer's IP address

int openRawSocket(const char *device)
{
	int sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (sock<0) { 
		if (errno == EPERM)
			panic("Cannot open raw socket on %s: Need to be root?", device);
		else 
			panic("Cannot open raw socket on %s (errno=%d)", device, errno);
	}

	struct ifreq ifr;
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));
	if (ioctl(sock, SIOCGIFINDEX, &ifr) < 0)
		panic("Cannot get the interface index of %s (errno=%d)", device, errno);

	struct sockaddr_ll ll;
	memset(&ll, 0, sizeof(ll));
	ll.sll_family = AF_PACKET;
	ll.sll_ifindex = ifr.ifr_ifindex;
	ll.sll_protocol = htons(ETH_P_ALL);
	if (bind(sock, (struct sockaddr*)&ll, sizeof(ll)) < 0)
		panic("Cannot bind to raw socket on %s (errno=%d)", device, errno);

	return sock;
}

void handleIncomingPacket(pcap_dumper_t *myDumper, unsigned char *buffer, int size, struct timeval *now)
{
	
	// Only log IP packets
	if((buffer[0] & 0xF0) == 0x40){
		int src, dst;
		memcpy((char *) &src, (char *) &buffer[12], 4);
		memcpy((char *) &dst, (char *) &buffer[16], 4);

		int csize = size;
		
		// Only log packets from/to the peer
		if(src == peer || dst == peer){
			
			int ip_hl = (buffer[0] & 0x0F)*4;
			// If UDP or TCP, strip packet content
			if(buffer[9] == IPPROTO_TCP){
				csize = ip_hl+((buffer[ip_hl+12] & 0xF0)/4);
				csize += 200;
				
				// If it is a command packet, log full packet content
				unsigned short src_port, dst_port;
				memcpy((char *) &src_port, (char *) &buffer[ip_hl], 2);
				memcpy((char *) &dst_port, (char *) &buffer[ip_hl+2], 2);
				 
				if((src_port == htons(command_port)) || (dst_port == htons(command_port)))
					csize = size;
			}
			else if(buffer[9] == IPPROTO_UDP){
				csize = ip_hl + 8;
				csize += 200;
			}
			if(csize > size)
				csize = size;
			
			// Anonymize IP header
			if(token != 0){
				if(src == peer){
					memcpy((char *) &buffer[12], (char *) &token, 4);
					memset((char *) &buffer[16], 0, 4);
				}
				else if(dst == peer){
					memcpy((char *) &buffer[16], (char *) &token, 4);
					memset((char *) &buffer[12], 0, 4);
				}
			}
			
			struct pcap_pkthdr hdr;
			hdr.ts = *now;
			hdr.caplen = csize;
			hdr.len = size;
			pcap_dump((u_char*)myDumper, &hdr, buffer);
			//  dumpPacket(buffer, size);
			
			if(buffer[9] == IPPROTO_TCP){
				// Check if we saw a RST and cache(?) it.			
				//int ip_hl = (buffer[0] & 0x0F)*4;
				if((src == peer) && ((buffer[ip_hl + 13] & 0x04))){
					// Got RST!!!
					//log("Got reset from peer!");
					resets += 1;
				}
				else if((dst == peer) && ((buffer[ip_hl + 13] & 0x04))){
					// Sent RST!!!
					//log("I sent a reset!");
					ownResets += 1;
				}
			}
		}
	}
}

/**
 * Calc the timeout to a given end time
 */
void calcTimeout(const struct timeval *tend, struct timeval *timeout){
	assert(timeout != NULL && tend != NULL);
	
	gettimeofday(timeout, NULL);
	timeout->tv_sec = tend->tv_sec - timeout->tv_sec;
	timeout->tv_usec = tend->tv_usec - timeout->tv_usec;

	if(timeout->tv_usec < 0){
		timeout->tv_usec += 1000000;
		timeout->tv_sec -= 1;
	}
	if(timeout->tv_sec < 0){
		timeout->tv_sec = 0;
		timeout->tv_usec = 0;
	}	
}

/**
 * Sniff packets until the timeout expires or the other socket is ready
 * to be read from.
 * 
 * @param sock socket to read from (e.g. a TCP or UDP socket)
 * @param read_socket (Raw) socket to sniff packets from
 * @param tend The end time of sniffing
 * 
 * @returns 0 if sock is ready to be read from or 1 if timeout seconds passed
 */
int sniffPackets(int sock, int read_socket, const struct timeval *tend){
	
	assert(sock >= 0 || read_socket >= 0);
	
	unsigned char pbuffer[65536];
	struct timeval now, tv;
	
	if(tend != NULL)
		calcTimeout(tend, &tv);
	
	if((tv.tv_sec == 0) && (tv.tv_usec == 0))
		return 1;
	
	while(true){
		
		fd_set fdr;
		FD_ZERO(&fdr);
		if(read_socket >= 0)
			FD_SET(read_socket, &fdr);
		if(sock >= 0)
			FD_SET(sock, &fdr);

		int sel = MAX(read_socket, sock);
		int numSel = 0;

		while ((numSel = select(sel+1, &fdr, NULL, NULL, &tv))<0)
			log("Select failed: errno=%d", errno); // TODO Might want to panic here
		
		// Timeout
		if(numSel == 0){
			return 1;
		}
		
		if(read_socket >= 0){
			if(FD_ISSET(read_socket, &fdr)){
				int size = read(read_socket, pbuffer, sizeof(pbuffer));
				gettimeofday(&now, NULL);
				if (size>14) {
					handleIncomingPacket(myDumper, &pbuffer[14], size - 14, &now);
				}
				// Adjust timeout if set
				if(tend != NULL){
					tv.tv_sec = tend->tv_sec - now.tv_sec;
					tv.tv_usec = tend->tv_usec - now.tv_usec;
					if(tv.tv_usec < 0){
						tv.tv_usec += 1000000;
						tv.tv_sec -= 1;
					}
					if(tv.tv_sec < 0){
						tv.tv_sec = 0;
						tv.tv_usec = 0;
					}	
				}
			}
		}
		if(sock >= 0){
			if(FD_ISSET(sock, &fdr))
				break;
		}
	}
	return 0;
}

int sniffPacketsForWrite(int sock, int read_socket, const struct timeval *tend){
	
	assert(sock >= 0 || read_socket >= 0);
	
	unsigned char pbuffer[65536];
	struct timeval now, tv;
	
	if(tend != NULL)
		calcTimeout(tend, &tv);
	
	if((tv.tv_sec == 0) && (tv.tv_usec == 0))
		return 1;
	
	while(true){
		
		fd_set fdr, fdw;
		FD_ZERO(&fdw);
		FD_ZERO(&fdr);
		if(read_socket >= 0)
			FD_SET(read_socket, &fdr);
		if(sock >= 0)
			FD_SET(sock, &fdw);

		int sel = MAX(read_socket, sock);
		int numSel = 0;

		while ((numSel = select(sel+1, &fdr, &fdw, NULL, &tv))<0)
			log("Select failed: errno=%d", errno);
		
		// Timeout
		if(numSel == 0){
			return 1;
		}
		
		if(read_socket >= 0){
			if(FD_ISSET(read_socket, &fdr)){
				int size = read(read_socket, pbuffer, sizeof(pbuffer));
				gettimeofday(&now, NULL);
				if (size>14) {
					handleIncomingPacket(myDumper, &pbuffer[14], size - 14, &now);
				}
				// Adjust timeout if set
				if(tend != NULL){
					tv.tv_sec = tend->tv_sec - now.tv_sec;
					tv.tv_usec = tend->tv_usec - now.tv_usec;
					if(tv.tv_usec < 0){
						tv.tv_usec += 1000000;
						tv.tv_sec -= 1;
					}
					if(tv.tv_sec < 0){
						tv.tv_sec = 0;
						tv.tv_usec = 0;
					}	
				}
			}
		}
		if(sock >= 0){
			if(FD_ISSET(sock, &fdw))
				break;
		}
	}
	return 0;
}

void handleSocketError(const char *message) {
	
	
	struct timeval tv;
	gettimeofday(&tv, NULL);
	tv.tv_sec += 2;
	sniffPackets(-1, read_socket, &tv);

	log("Resets: %d ; Resets sent: %d", resets, ownResets);
	
	log("%s", message);
}

long long getTimeMicros()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec * 1000000LL + tv.tv_usec;
}

const int readBufSizeBytes = pieceSize + 16*1024;
const int writeBufSizeBytes = pieceSize + 16*1024;
const int progressReportAfterBytes = 100000;
unsigned char readBuf[readBufSizeBytes];
unsigned char writeBuf[writeBufSizeBytes];
unsigned char rawBuf[65536];
int readLen;
int writePtr;
int writeLen;

void chompRead(int numBytes)
{
  assert(numBytes <= readLen);
  
  for (int i=numBytes; i<readLen; i++)
    readBuf[i-numBytes] = readBuf[i];
    
  readLen -= numBytes;
}

#define final const
#define byte unsigned char

bool checkString(unsigned char *buf, int offset, const char *str)
{
  int i = 0;
  while (true) {
    if (str[i] == 0)
      return true;
    
    if (buf[i+offset] != str[i])
      return false;
      
    i++;
  }
}

void putString(unsigned char *buf, int offset, const char *str)
{
  for (int i=0; str[i]; i++)
    buf[i+offset] = str[i];
}

unsigned char getTrulyRandomByte()
{
  return rand() & 0xFF;
}

unsigned char getTrulyRandomInt()
{
  return rand();
}

void fillWithTrulyRandomBytes(unsigned char *buf, int offset, int len)
{
  for (int i=0; i<len; i++)
    buf[offset+i] = (rand()&0xFF);
}

void fillWithPseudoRandomBytes(unsigned char *buf, int offset, int len)
{
  const int randomnessSize = 100;
  unsigned char randomness[randomnessSize] = { 
    0x67, 0xC6, 0x69, 0x73, 0x51, 0xFF, 0x4A, 0xEC, 0x29, 0xCD, 
    0xBA, 0xAB, 0xF2, 0xFB, 0xE3, 0x46, 0x7C, 0xC2, 0x54, 0xF8, 
    0x1B, 0xE8, 0xE7, 0x8D, 0x76, 0x5A, 0x2E, 0x63, 0x33, 0x9F, 
    0xC9, 0x9A, 0x66, 0x32, 0x0D, 0xB7, 0x31, 0x58, 0xA3, 0x5A, 
    0x25, 0x5D, 0x05, 0x17, 0x58, 0xE9, 0x5E, 0xD4, 0xAB, 0xB2, 
    0xCD, 0xC6, 0x9B, 0xB4, 0x54, 0x11, 0x0E, 0x82, 0x74, 0x41, 
    0x21, 0x3D, 0xDC, 0x87, 0x70, 0xE9, 0x3E, 0xA1, 0x41, 0xE1, 
    0xFC, 0x67, 0x3E, 0x01, 0x7E, 0x97, 0xEA, 0xDC, 0x6B, 0x96, 
    0x8F, 0x38, 0x5C, 0x2A, 0xEC, 0xB0, 0x3B, 0xFB, 0x32, 0xAF, 
    0x3C, 0x54, 0xEC, 0x18, 0xDB, 0x5C, 0x02, 0x1A, 0xFE, 0x43
  };

  for (int i=0; i<len; i++)
    buf[offset+i] = randomness[i%randomnessSize];
}

void putInt(unsigned char *buf, int offset, int value)
{
  *(int*)&buf[offset] = htonl(value);
}

int getInt(unsigned char *buf, int offset)
{
  return ntohl(*(int*)&buf[offset]);
}

void dump(FILE *stream, const unsigned char *payload, int len, int indent=0)
{
  int i, off = 0;
  while (off < len) {
    for (int k=0; k<indent; k++)
      fprintf(stream, " ");
      
    fprintf(stream, "%04X   ", off);
    for (i=0; i<16; i++) {
      if ((i+off) < len)
        fprintf(stream, "%02X ", payload[i+off]);
      else
        fprintf(stream, "   ");
    }
    
    fprintf(stream, "   ");
    
    for (i=0; i<16; i++) {
      if ((i+off) < len)
        fprintf(stream, "%c", isprint(payload[i+off]) ? payload[i+off] : '.');
    }
    
    off += 16;
    fprintf(stream, "\n");
  }
}

/* This function implements four different types of streams: 
       - a BitTorrent seeder
       - a BitTorrent leecher
       - a TCP sender that uses messages of the same sizes as the BitTorrent seeder, 
           but with random content, and
       - a TCP receiver that uses messages of the same sizes as the BitTorrent leecher, 
           but with random content
   The caller must provide the file descriptor of a socket that is already connected to 
   the other endpoint, plus the overall duration of the upload/download in seconds. */

double connectionHandler(int fd, int durationSec, bool useRealProtocol, bool isSeeder, struct timeval *start, struct timeval *end)
{
  assert((fd >= 0) && (read_socket >= 0));

  gettimeofday(start, NULL);
  
  long long endTime = getTimeMicros() + (durationSec * 1000000LL);       // expected end of transfer
  int state = isSeeder ? 0x00 : 0x10;            // our current protocol state (see STATE_XXX below)
  int statLastIndex = 0;                      // the read/write pointer after the last read or write
  int statBytesTransferred = 0;     // total number of payload bytes transferred so far (no headers)
  int statPiecesTransferred = 0;             // total number of BitTorrent pieces transferred so far
    
  /* Initialize some variables */
    
  readLen = 0;
  writePtr = 0;
  writeLen = 0;
  transferSuccessful = 0;
  
  /* These two variables keep a) the total number of RST packets we've seen, and b) the number
     of RST packets we've sent locally. They're updated by handleIncomingPacket(), and the
     result is output at the end of this function. */
  
  resets = 0; 
  ownResets = 0;

  bool protocolError = false;
  bool connectionClosed = false;
    
  /* Main loop */
    
  while (!protocolError && !connectionClosed) {

		/**********************************************************************************************************
		 * BitTorrent state machine. This is also valid Java code, and can be copied 1:1 to                       *
		 * BitTorrentClient.java                                                                                  */

		// The states of the BitTorrent protocol, separated by Seeder (start at 0x00) and Leecher (start at 0x10)
		final int STATE_SEEDER_WAIT_HANDSHAKE = 0x00;
		final int STATE_SEEDER_SEND_HANDSHAKE = 0x01;
		final int STATE_SEEDER_WAIT_BITFIELD = 0x02;
		final int STATE_SEEDER_SEND_BITFIELD = 0x03;
		final int STATE_SEEDER_WAIT_INTERESTED = 0x04;
		final int STATE_SEEDER_SEND_UNCHOKE = 0x05;
		final int STATE_SEEDER_WAIT_REQUEST_OR_CANCEL = 0x06;
		final int STATE_SEEDER_SEND_PIECE = 0x07;
		final int STATE_LEECHER_INIT_HANDSHAKE = 0x10;
		final int STATE_LEECHER_SEND_HANDSHAKE = 0x11;
		final int STATE_LEECHER_WAIT_HANDSHAKE = 0x12;
		final int STATE_LEECHER_SEND_BITFIELD = 0x13;
		final int STATE_LEECHER_WAIT_BITFIELD = 0x14;
		final int STATE_LEECHER_SEND_INTERESTED = 0x15;
		final int STATE_LEECHER_WAIT_UNCHOKE = 0x16;
		final int STATE_LEECHER_SEND_REQUEST = 0x17;
		final int STATE_LEECHER_WAIT_PIECE = 0x18;

		switch (state) {
		case STATE_LEECHER_INIT_HANDSHAKE: {
			/* The leecher starts by sending a handshake message. We make up random hash and ID values */

			if (useRealProtocol) {
				for (int i = 0; i < 68; i++)
					writeBuf[i] = 0;
				writeBuf[0] = 19;
				putString(writeBuf, 1, "BitTorrent protocol");
				for (int i = 0; i < 20; i++)
					writeBuf[28 + i] = getTrulyRandomByte(); // hash value
				for (int i = 0; i < 20; i++)
					writeBuf[48 + i] = getTrulyRandomByte(); // id value
			} else {
				fillWithTrulyRandomBytes(writeBuf, 0, 68);
			}

			writeLen = 68;
			writePtr = 0;

			log2("Sending handshake message");
			state = STATE_LEECHER_SEND_HANDSHAKE;

			break;
		}
		case STATE_LEECHER_SEND_HANDSHAKE: {
			/* We wait until the entire message has been sent */

			if (writePtr == writeLen) {
				writePtr = writeLen = 0;

				log2("Waiting for handshake message");
				state = STATE_LEECHER_WAIT_HANDSHAKE;
			}

			break;
		}
		case STATE_LEECHER_WAIT_HANDSHAKE: /* Waits for a BitTorrent handshake message */
		{
			/* Now we wait for a handshake message from the Seeder. Once we receive one, we perform some 
			sanity checks. Since we know exactly which programm is running on the other side, this
			check should never fail. If it does, something is seriously wrong! */

			if (readLen >= 68) {
				if (useRealProtocol	&& ((readBuf[0] != 19) || !checkString(readBuf,	1, "BitTorrent protocol"))) {
					log2("Packet is not a BT handshake");
					protocolError = true;
					break;
				}

				log2("Received handshake message");
				chompRead(68);

				/* We respond with a bitfield message. We claim that we have nothing at all. */

				if (useRealProtocol) {
					for (int i = 0; i < 166; i++)
						writeBuf[i] = 0x00;
					
					putInt(writeBuf, 0, 162); // Msg length
					writeBuf[4] = 5;  // msg type
				} else {
					fillWithTrulyRandomBytes(writeBuf, 0, 166);
				}

				writeLen = 166;
				writePtr = 0;

				log2("Sending bitfield message");
				state = STATE_LEECHER_SEND_BITFIELD;
			}

			break;
		}
		case STATE_LEECHER_SEND_BITFIELD: {
			/* We wait until the entire message has been sent */

			if (writePtr == writeLen) {
				writePtr = writeLen = 0;

				log2("Waiting for bitfield message");
				state = STATE_LEECHER_WAIT_BITFIELD;
			}

			break;
		}
		case STATE_LEECHER_WAIT_BITFIELD: {
			/* Now we wait for a bitfield message from the Seeder. */

			if (readLen >= 166) {
				if (useRealProtocol	&& (readBuf[4] != 5)) {
					log2("Packet is not a BT bitfield message");
					protocolError = true;
					break;
				}

				log2("Received bitfield message");
				chompRead(166);

				/* We respond with an 'interested' message. */

				if (useRealProtocol) {
					putInt(writeBuf, 0, 1);
					writeBuf[4] = 2;
				} else {
					fillWithTrulyRandomBytes(writeBuf, 0, 5);
				}

				writeLen = 5;
				writePtr = 0;

				log2("Sending interested message");
				state = STATE_LEECHER_SEND_INTERESTED;
			}

			break;
		}
		case STATE_LEECHER_SEND_INTERESTED: {
			/* We wait until the entire message has been sent */

			if (writePtr == writeLen) {
				writePtr = writeLen = 0;

				log2("Waiting for unchoke message");
				state = STATE_LEECHER_WAIT_UNCHOKE;
			}

			break;
		}
		case STATE_LEECHER_WAIT_UNCHOKE: {
			/* We wait for an 'unchoke' message from the Seeder */

			if (readLen >= 5) {
				if (useRealProtocol	&& (readBuf[4] != 1)) {
					log2("Packet is not a BT unchoke message");
					protocolError = true;
					break;
				}

				log2("Received unchoke message");
				chompRead(5);

				/* We respond with a request message. Note that if we're using random content instead
				of real BitTorrent messages, the Seeder cannot distinguish between a REQUEST and
				a CANCEL. Therefore we set the 12th byte to 1 for REQUESTs, and to 0 for CANCELs.  */						
				if (useRealProtocol) {
					putInt(writeBuf, 0, 13);
					writeBuf[4] = 6;							
					putInt(writeBuf, 5, getTrulyRandomInt() % 1345); // piece index
					putInt(writeBuf, 9, 0); // offset
					putInt(writeBuf, 13, pieceSize); // length
				} else {
					fillWithTrulyRandomBytes(writeBuf, 0, 17);
					writeBuf[11] = 1;
				}

				writeLen = 17;
				writePtr = 0;

				log2("Sending request message");
				state = STATE_LEECHER_SEND_REQUEST;
			}

			break;
		}
		case STATE_LEECHER_SEND_REQUEST: {
			/* We wait until the entire message has been sent */

			if (writePtr == writeLen) {
				writePtr = writeLen = 0;
				statLastIndex = 0;

				//log2("Waiting for piece message");
				state = STATE_LEECHER_WAIT_PIECE;
			}

			break;
		}
		case STATE_LEECHER_WAIT_PIECE: {
			/* Now we're downloading a PIECE message. While the message is not yet complete, we 
			update our statistics in case the conncetion is aborted later. Also, we output
			progress reports periodically. */

			if (readLen > statLastIndex) {
				int newStatBytesTransferred = statBytesTransferred + (readLen - statLastIndex);
				//if (((int) (newStatBytesTransferred / progressReportAfterBytes)) != ((int) (statBytesTransferred / progressReportAfterBytes)))
				//	log2i("Received so far: ", newStatBytesTransferred);

				statBytesTransferred = newStatBytesTransferred;
				statLastIndex = readLen;
			}

			/* Once the message is complete, we parse it */

			if (readLen >= (pieceSize + 13)) {
				if (useRealProtocol && (readBuf[4] != 7)) {
					log2("Packet is not a BT piece message");
					protocolError = true;
					break;
				}

				//log2("Received piece message");
				chompRead(pieceSize + 13);
				statPiecesTransferred++;

				/* And then we send another request message */

				if (useRealProtocol) {
					putInt(writeBuf, 0, 13);
					writeBuf[4] = 6;							
					putInt(writeBuf, 5, getTrulyRandomInt() % 1345); // piece index
					putInt(writeBuf, 9, 0); // offset
					putInt(writeBuf, 13, pieceSize); // length
				} else {
					fillWithTrulyRandomBytes(writeBuf, 0, 17);
					writeBuf[11] = 1;
				}

				writeLen = 17;
				writePtr = 0;

				//log2("Sending request message");
				state = STATE_LEECHER_SEND_REQUEST;
			}

			break;
		}

		/* BitTorrent Seeder ---------------------------------------------------------------------------------*/

		case STATE_SEEDER_WAIT_HANDSHAKE: {
			/* The Seeder initially waits for a handshake message from the Leecher. */

			if (readLen >= 68) {
				if (useRealProtocol	&& ((readBuf[0] != 19) || !checkString(readBuf,	1, "BitTorrent protocol"))) {
					log2("Packet is not a BT handshake");
					protocolError = true;
					break;
				}

				log2("Received handshake message");
				chompRead(68);

				/* We respond with a handshake message of our own */

				if (useRealProtocol) {
					for (int i = 0; i < 68; i++)
						writeBuf[i] = 0;
					writeBuf[0] = 19;
					putString(writeBuf, 1, "BitTorrent protocol");
					for (int i = 0; i < 20; i++)
						writeBuf[28 + i] = readBuf[28 + i]; // the hash the Leecher provided
					for (int i = 0; i < 20; i++)
						writeBuf[48 + i] = getTrulyRandomByte(); // random ID
				} else {
					fillWithTrulyRandomBytes(writeBuf, 0, 68);
				}

				writeLen = 68;
				writePtr = 0;

				log2("Sending handshake message");
				state = STATE_SEEDER_SEND_HANDSHAKE;
			}

			break;
		}
		case STATE_SEEDER_SEND_HANDSHAKE: {
			/* We wait until the entire message has been sent */

			if (writePtr == writeLen) {
				writePtr = writeLen = 0;

				log2("Waiting for bitfield message");
				state = STATE_SEEDER_WAIT_BITFIELD;
			}

			break;
		}
		case STATE_SEEDER_WAIT_BITFIELD: {
			/* We wait for a bitfield message */

			if (readLen >= 166) {
				if (useRealProtocol && (readBuf[4] != 5)) {
					log2("Packet is not a BT bitfield message");
					protocolError = true;
					break;
				}

				log2("Received bitfield message");
				chompRead(166);

				/* We respond with our own bitfield message. Since we're the Seeder, we claim
				that we have everything (bitfield is all ones) */

				if (useRealProtocol) {
					for (int i = 0; i < 166; i++)
						writeBuf[i] = (byte) 0xFF;

					putInt(writeBuf, 0, 162); // Msg length
					writeBuf[4] = 5;  // msg type
				} else {
					fillWithTrulyRandomBytes(writeBuf, 0, 166);
				}

				writeLen = 166;
				writePtr = 0;

				log2("Sending bitfield message");
				state = STATE_SEEDER_SEND_BITFIELD;
			}

			break;
		}
		case STATE_SEEDER_SEND_BITFIELD: {
			/* We wait until the entire message has been sent */

			if (writePtr == writeLen) {
				writePtr = writeLen = 0;

				log2("Waiting for interested message");
				state = STATE_SEEDER_WAIT_INTERESTED;
			}

			break;
		}
		case STATE_SEEDER_WAIT_INTERESTED: {
			/* We wait for an 'interested' message from the Leecher */

			if (readLen >= 5) {
				if (useRealProtocol	&& (readBuf[4] != 2)) {
					log2("Packet is not a BT interested message");
					protocolError = true;
					break;
				}

				log2("Received interested message");
				chompRead(5);

				/* We respond with an unchoke message */

				if (useRealProtocol) {
					putInt(writeBuf, 0, 1);
					writeBuf[4] = 1;						
				} else {
					fillWithTrulyRandomBytes(writeBuf, 0, 5);
				}

				writeLen = 5;
				writePtr = 0;

				log2("Sending unchoke message");
				state = STATE_SEEDER_SEND_UNCHOKE;
			}

			break;
		}
		case STATE_SEEDER_SEND_UNCHOKE: {
			/* We wait until the entire message has been sent */

			if (writePtr == writeLen) {
				writePtr = writeLen = 0;

				//log2("Waiting for request or cancel message");
				state = STATE_SEEDER_WAIT_REQUEST_OR_CANCEL;
			}

			break;
		}
		case STATE_SEEDER_WAIT_REQUEST_OR_CANCEL: {
			/* At this point, the next message could either be a REQUEST or a CANCEL.
			If we're using the randomized protocol, we distinguish between the two
			using the 12th byte. */
			
			if ((readLen >= 17)	&& (useRealProtocol ? (readBuf[4] == 8) : (readBuf[11] == 0))) {
				log2("Received cancel message");
				connectionClosed = true;
			} else if ((readLen >= 17) && (useRealProtocol ? (readBuf[4] == 6) : (readBuf[11] == 1))) {
				//log2("Received request message");
				chompRead(17);

				if (useRealProtocol) {				
				
					/* Extract parameters from the REQUEST message */

					int blength = getInt(readBuf, 13);
					int index = getInt(readBuf, 5);
					int offset = getInt(readBuf, 9);

					/* Perform some sanity checks */

					if (blength > pieceSize)
						blength = pieceSize;
					if (blength < 0)
						blength = pieceSize;

					/* Make a piece message. */

					putInt(writeBuf, 0, 9 + pieceSize);
					writeBuf[4] = 7;
					putInt(writeBuf, 5, index);
					putInt(writeBuf, 9, offset);

					fillWithPseudoRandomBytes(writeBuf, 13, blength);
					writeLen = blength + 13;
				}
				else {
					fillWithPseudoRandomBytes(writeBuf, 0, pieceSize + 13);
					writeLen = pieceSize + 13;
				}							
				
				writePtr = 0;
				statLastIndex = 13;

				//log2("Sending piece message");
				state = STATE_SEEDER_SEND_PIECE;
			} else if (readLen >= 17) {
				log2("Packet is neither a BT request nor a BT cancel");
				protocolError = true;
				break;
			}

			break;
		}
		case STATE_SEEDER_SEND_PIECE: {
			/* We wait until the entire message has been sent. Again, we keep our statistics updated
			in case the connection is suddenly aborted. */

			if (writePtr > statLastIndex) {
				int newStatBytesTransferred = statBytesTransferred + (writePtr - statLastIndex);
				//if (((int) (newStatBytesTransferred / progressReportAfterBytes)) != ((int) (statBytesTransferred / progressReportAfterBytes)))
				//	log2i("Sent so far: ", newStatBytesTransferred);

				statBytesTransferred = newStatBytesTransferred;
				statLastIndex = writePtr;
			}

			if (writePtr == writeLen) {
				writePtr = writeLen = 0;
				statPiecesTransferred++;

				//log2("Waiting for next request or cancel message");
				state = STATE_SEEDER_WAIT_REQUEST_OR_CANCEL;
			}

			break;
		}
		default: {
			log2("Unknown protocol state");
			protocolError = true;
			break;
		}
		}

		/* End of the BitTorrent state machine. The code below is no longer portable.                  *
		/***********************************************************************************************/
  
    /* If something has gone wrong, abort immediately (otherwise we block at the end) */
  
    if (protocolError || connectionClosed)
      break;
  
    /* Also, abort if the transfer has timed out. This is the 'normal' way to exit, which
       is taken if the connection is not aborted or otherwise interfered with. */
  
    long long now = getTimeMicros();
    if (now >= endTime)
      break;
      
    /* Prepare the bitfields for select(). We always want to read packets from the raw socket 
       (so that we can dump them to the pcap trace), and we almost always want to read data,
       except if our buffer is full. We only want to write if we have something to send. */
      
    fd_set fdr, fdw;
    int maxSocket = 0;
    FD_ZERO(&fdr);
    FD_ZERO(&fdw);
    FD_SET(read_socket, &fdr);
    if (read_socket > maxSocket)
      maxSocket = read_socket;
    if (readLen < readBufSizeBytes) {
      FD_SET(fd, &fdr);
      if (fd > maxSocket)
        maxSocket = fd;
    }
    if (writePtr < writeLen) {
      FD_SET(fd, &fdw);
      if (fd > maxSocket)
        maxSocket = fd;
    }
     
    /* Calculate the timeout value */
      
    struct timeval tv;
    tv.tv_sec = (int)((endTime-now) / 1000000LL);
    tv.tv_usec = (int)((endTime-now) % 1000000LL);
  
    /* Call select. If we're interrupted in the middle (signal etc.), we continue. 
       TODO: Examine errno and panic if appropriate (to prevent spinning) */
  
    do {
      int ret = select(maxSocket+1, &fdr, &fdw, NULL, &tv);
      if (ret < 0) {
        log2x("Select failed (errno=%d, state=%d)", errno, state);
        continue;
      }
    } while (false);
  
    /* If we got a packet on the raw socket, we check if it belongs to this connection,
       then we write it to the pcap trace file */
  
    if (FD_ISSET(read_socket, &fdr)) {
      int size = read(read_socket, rawBuf, sizeof(rawBuf));
      struct timeval tnow;
      gettimeofday(&tnow, NULL);
      if (size>14) 
        handleIncomingPacket(myDumper, &rawBuf[14], size - 14, &tnow);
    }
    
    /* If the connection is ready for reading, we do that */
    
    if (FD_ISSET(fd, &fdr)) {
      int len = read(fd, &readBuf[readLen], readBufSizeBytes - readLen);
      if (len < 0) {
        log2x("Read failed (errno=%d, state=%d)", errno, state);
        connectionClosed = true;
        break;
      }
      
      if (len == 0) {
        log2x("Connection closed (read, errno=%d, state=%d)", errno, state);
        connectionClosed = true;
        break;
      }
      
      //dump(stdout, &readBuf[readLen], len);
      readLen += len;
    }

    /* If the connection is ready for writing, we do that */
    
    if (FD_ISSET(fd, &fdw)) {
      assert(writePtr < writeLen);
      int len = write(fd, &writeBuf[writePtr], writeLen - writePtr);
      //dump(stdout, &writeBuf[writePtr], writeLen-writePtr);
      
      if (len < 0) {
        log2x("Write failed (errno=%d, state=%d)", errno, state);
        connectionClosed = true;
        break;
      }
      if (len == 0) {
        //log2x("Connection closed (write, errno=%d, state=%d)", errno, state);
        connectionClosed = true;
        break;
      }
      
      writePtr += len;
    }
  }

  gettimeofday(end, NULL);
  
  /* End of the main loop. We could have reached this point in several ways: The connection
     could have timed out (yippie!), it could have been interrupted by the ISP (gotcha!),
     or we could have run into a protocol error (bummer!) */

  log2x("End of transfer; %d bytes and %d pieces transferred", statBytesTransferred, statPiecesTransferred);

  /* We can't return immediately because we might otherwise miss the RST from the other
     side. To ensure the RST makes it to the pcap trace, we spend two additional seconds
     capturing packets. */
  
  //log2("Spending an additional two seconds sniffing packets");
  struct timeval tv;
  gettimeofday(&tv, NULL);
  tv.tv_sec += 2;
  sniffPackets(-1, read_socket, &tv);

  
  log("Resets: %d ; Resets sent: %d", resets, ownResets);

  transferSuccessful = (!protocolError && (statBytesTransferred > 10000));
  
  return statBytesTransferred;
}

double tcpReceiver(int fd, int duration){
	
	resets = 0; ownResets = 0;
	transferSuccessful = 1;
	char buffer[4096];
	struct timeval tend;
	gettimeofday(&tend, NULL);
	tend.tv_sec = tend.tv_sec + duration;
	isTerminated = 0;
	
	double dataReceived = 0;
	while(!isTerminated){

		gettimeofday(&tend, NULL);
		tend.tv_sec = tend.tv_sec + duration;
		
		if(sniffPackets(fd, read_socket, &tend) == 1)
			break;
 
		int ret = read(fd, buffer, sizeof(buffer));
		if(ret == -1){
			handleSocketError("tcpReceiver: Cannot read from socket.");
			if(dataReceived <= 0)
				transferSuccessful = 0;
			isTerminated = 1;
			break;
		}
		if(ret == 0)
			break;
		
		dataReceived += ret;
	}
	return dataReceived;
}

double tcpSender(int fd, int duration){
	
	resets = 0; ownResets = 0;
	transferSuccessful = 1;
	char buffer[16000];
	struct timeval tend;
	gettimeofday(&tend, NULL);
	tend.tv_sec = tend.tv_sec + duration;
	isTerminated = 0;
	srand(tend.tv_sec);
	
	double dataSent = 0;
	while(!isTerminated){

		for(int i=0; i<sizeof(buffer); i++)
			buffer[i] = (char) rand();
		
		if(sniffPacketsForWrite(fd, read_socket, &tend) == 1)
			break;
		
		int ret = write(fd, buffer, sizeof(buffer));
		if(ret < 0 /* && errno != EAGAIN */){
			handleSocketError("tcpSender: Cannot write to socket.");
			if(dataSent <= 0)
				transferSuccessful = 0;
			isTerminated = 1;
			break;
		}
		if(ret == 0)
			break;
		
		dataSent += ret;
	}
	return dataSent;
}

void setNonblocking(int fd)
{
  int flags = fcntl(fd, F_GETFL, &flags);
  if (flags>=0) {
    flags |= O_NONBLOCK;

    int err = fcntl(fd, F_SETFL, flags);
    if (err<0)
      log2x("Cannot set socket %d to nonblocking (errno=%d)", fd, errno);
    else
      log2x("Socket %d set to nonblocking", fd);
  } else {
    log2x("Cannot get socket flags for %d (errno=%d)", fd, errno);
  }
}


/**
 * What parameter main() takes. 
 * -z zips the dump file
 */
void usage(){
	panic("Usage: bt_client [-p <commandPort>] [-h <peerIP>] [-i ethDev] [-o dumpfile] [-a anonymizationToken] [-z]");
}

int main(int argc, char *args[]){
	if(argc > 12){
		usage();
	}

	char eth_dev[200];
	strcpy(eth_dev, "eth0");
	char buffer[4096];
	bool zip = false;
	
	struct sockaddr_in addr;
	
	for (int i=1; i<argc; i++) {
		if (!strcmp(args[i], "-p")) {
			i++;
			if(i >= argc) usage();
			command_port = atoi(args[i]);
		}
		else if (!strcmp(args[i], "-h")) {
			i++;
			if(i >= argc) usage();
						
			struct hostent *hp;
			if ((hp=gethostbyname(args[i])) == NULL)
				panic("Error: Cannot find host %s", args[i]);
			memcpy(&peer, hp->h_addr, MIN(sizeof(int), hp->h_length));
		}
		else if (!strcmp(args[i], "-i")){
			i++;
			if(i >= argc) usage();
			strcpy(eth_dev, args[i]);			
		}
		else if (!strcmp(args[i], "-o")){
			i++;
			if(i >= argc) usage();
			strcpy(dumpFile, args[i]);			
		}
		else if (!strcmp(args[i], "-a")){
			i++;
			if(i >= argc) usage();
			token = atoi(args[i]);			
		}
		else if (!strcmp(args[i], "-z")){
			zip = true;
		}
		else {
			fprintf(stderr, "Unknown option: %s\n", args[i]);
			usage();
		}
	}		
	
	fprintf(stderr, "bt_client: Starting up.\n");
        log2x("Starting BT client v%s", VERSION);
        log2x("Configuration: -p %d -h %d.%d.%d.%d -i %s -o %s", command_port, peer&0xFF, (peer>>8)&0xFF, (peer>>16)&0xFF, (peer>>24)&0xFF, eth_dev, dumpFile);
	
        
	signal(SIGPIPE, SIG_IGN);

	
	struct timeval tv;
	int cs; // Accepted command connection
	
	while(1){
		/* Get command socket */
		if ((command_sock=socket(AF_INET, SOCK_STREAM, 0))<0)
			panic("Cannot create socket\n");

		int on = 1;
		setsockopt(command_sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

		/* Bind to the given port */
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = INADDR_ANY;
		addr.sin_port = htons(command_port);

		if (bind(command_sock, (struct sockaddr*)&addr, sizeof(addr)) < 0){
			perror("Error");
			panic("bind() command socket failed.");
		}

		/* Set it to listen */
		if (listen(command_sock, 1) < 0)
			panic("listen() on command socket failed\n");

		gettimeofday(&tv, NULL);
		tv.tv_sec += 30;
		
		if(sniffPackets(command_sock, read_socket, &tv) == 1)
			panic("Accept timeout: Peer did not show up to give me commands.");

		/* Accept the incoming connection */
		struct sockaddr_in from;
		int len = sizeof(from);
		cs = accept(command_sock, (struct sockaddr*)&from, (socklen_t*)&len);
		if (cs < 0)
			panic("Cannot accept new connection: %d", errno);

		// Check if the connecting host is the one we are interested in
		if((peer != -1) && (from.sin_addr.s_addr != peer)){
			log("Connection from %s not accepted as we filter for another host.", inet_ntoa(from.sin_addr));
			close(cs);
			continue;
		}
		else if(peer == -1){
			peer = from.sin_addr.s_addr;
		}
		//log("Client connected on command port.");
				
		break;
	}

	char peerName[500];
	memset(&peerName, 0, 500);
	struct hostent *hp = gethostbyaddr((char *)&peer, sizeof(int), AF_INET);
	strncpy((char *) &peerName, (hp ? hp->h_name : inet_ntoa(*(struct in_addr*)&peer)), 499);
	log("Client %s %s connected", peerName, inet_ntoa(*(struct in_addr*)&peer));


	if(! strcmp(dumpFile, "") || strcmp(dumpFile, "auto")){
		gettimeofday(&tv, NULL);
		snprintf(dumpFile, 500, "bt_%s_%s_%d.dump", inet_ntoa(*(struct in_addr*)&peer), peerName, (int)tv.tv_sec);
		log("Saving trace to %s", dumpFile);
		
		//	strcpy(dumpFile, "bt.dump");
		
	}
	
	// Make the sniffing (raw) socket non-blocking 
	read_socket = openRawSocket(eth_dev);
	int flags = fcntl(read_socket, F_GETFL, &flags);
	if (flags<0)
		panic("Cannot F_GETFL the raw socket");
	flags |= O_NONBLOCK;

	int err = fcntl(read_socket, F_SETFL, flags);
	if (err<0)
		panic("Cannot F_SETFL the raw socket (errno=%d)", errno);

	myPcap = pcap_open_dead(DLT_RAW, 4096);
	if (!myPcap)
		panic("Cannot initialize libpcap");

	myDumper = pcap_dump_open(myPcap, dumpFile);
	if (!myDumper)
		panic("Cannot create pcap trace: '%s'", dumpFile); 		

	
	double bytesTransmitted = 0;
	
	char cbuff[2000];
	int cbuff_p = 0;
	
	bool connectionTerminated = false;
	
	while(!connectionTerminated){

		bool commandFound = false;

		while((!connectionTerminated) && (!commandFound)){
			for(int i=0; (i < cbuff_p) && (!commandFound); i++){
				if(cbuff[i] == '\n'){
					for(int n=0; n <= i; n++)
						buffer[n] = cbuff[n];
					buffer[i] = 0;

					for(int n=i+1; n < cbuff_p; n++)
						cbuff[n-(i+1)] = cbuff[n];

					cbuff_p -= (i+1);

					commandFound = true;
				}
			}

			if(!commandFound){
				// Read from command_socket
				assert(cbuff_p < sizeof(cbuff));

				gettimeofday(&tv, NULL);
				tv.tv_sec += 10;
				if(sniffPackets(cs, read_socket, &tv) == 1)
					panic("Peer did not give me commands.");

				int ret = read(cs, &cbuff[cbuff_p], sizeof(cbuff) - cbuff_p);
				if(ret == -1)
					panic("Failed to read commands.");

				if(ret == 0)
					connectionTerminated = true;
				
				cbuff_p += ret;
			}
		}
		
		if(connectionTerminated)
			break;
		assert(commandFound);
			
			
		int port, duration, exp, direction;
		if (!strncmp(buffer, "setup", 5)) {
			exp = SETUP;
		}
		else if(!strncmp(buffer, "bt", 2)) {
			exp = BT;
		}
		else if(!strncmp(buffer, "tcp", 3)) {
			exp = TCP;
		}
		else if(!strncmp(buffer, "shutdown", 8)) {
			connectionTerminated = true;
			break;
		}
		else if(!strncmp(buffer, "log", 3)) {			
			fprintf(stdout, "# Client: %s\n", (char *)(&buffer[4]));
			continue;
		}
		else{
			fprintf(stderr, "Unknown command: %s\n", buffer);
			continue;
		}

		char *arg[20];
		arg[0] = strtok(buffer, " ");
		for (int i=1; i<20; i++)
			arg[i] = arg[i-1] ? strtok(NULL, " ") : NULL;
		
		if(!strcmp(arg[1], "upstream")){
			direction = UPSTREAM;
		}
		else if(!strcmp(arg[1], "downstream")){
			direction = DOWNSTREAM;
		}
		else{
			fprintf(stderr, "Malformed command: %s\n", buffer);
			continue;
		}

		duration = atoi(arg[2]);
		if(strcmp(arg[3], "port")){
			fprintf(stderr, "Malformed command: %s\n", buffer);
			continue;
		}
		port = atoi(arg[4]);
		
		printf("Received: ");
		if(exp == SETUP) printf("setup ");
		else if(exp == BT) printf("bt ");
		else if(exp == TCP) printf("tcp ");
		if(direction == UPSTREAM) printf("upstream ");
		else if(direction == DOWNSTREAM) printf("downstream ");
		printf("port %d duration %d\n", port, duration);
		
		
		// Now start the experiment
		struct timeval start, end;
		bytesTransmitted = 0;
	
		/* Get a socket */
		if ((sock=socket(AF_INET, SOCK_STREAM, 0))<0)
			panic("Cannot create socket\n");

		int on = 1;
		setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

		/* Bind to the given port */
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = INADDR_ANY;
		addr.sin_port = htons(port);
		
		if (bind(sock, (struct sockaddr*)&addr, sizeof(addr)) < 0){
			perror("Error");
			panic("bind() failed for port %d", port);
		}
		
		// Set it to listen
		if (listen(sock, 1) < 0)
			panic("listen() failed\n");

		// Send back an "ok\n" to the server
		int ret = write(cs, "ok\n", 3);
		if(ret < 3)
			panic("Cannot write to command socket.");		
		
		int fd;
		while(1){
			// Wait for at most 30 seconds for the client to connect
			gettimeofday(&tv, NULL);
			tv.tv_sec += 30;
			if(sniffPackets(sock, read_socket, &tv) == 1)
				panic("Accept timeout: Peer did not show up.");

			// Accept the incoming connection
			struct sockaddr_in from;
			int len = sizeof(from);
			fd = accept(sock, (struct sockaddr*)&from, (socklen_t*)&len);
			if (fd < 0)
				panic("Cannot accept new connection: %d", errno);
			
			// Check if the connecting host is the one we are interested in
			if((peer != -1) && (from.sin_addr.s_addr != peer)){
				log("Connection from %s not accepted as we filter for another host.", inet_ntoa(from.sin_addr));
				close(fd);
							
				continue;
			}
			
			break;
		}
				
		if((exp == SETUP) && (direction == DOWNSTREAM)){
			log("Setting up downstream transfer.");
			tcpReceiver(fd, duration);
		}
		else if((exp == BT) && (direction == DOWNSTREAM)){
			log("Running BT downstream transfer port %d .", port);
			setNonblocking(fd);
			bytesTransmitted = connectionHandler(fd, duration, true, false, &start, &end);
			
			double timespan;
			if((end.tv_usec - start.tv_usec) >= 0)
				timespan = (end.tv_sec - start.tv_sec) + ((end.tv_usec - start.tv_usec)/1000000.0);
			else
				timespan = (end.tv_sec - start.tv_sec -1) + ((1000000 + end.tv_usec - start.tv_usec)/1000000.0);
			
			if(transferSuccessful)
				log("Transfered %.0f bytes in %.3f seconds: %.0f bps", bytesTransmitted, timespan, (bytesTransmitted*8/timespan));
			else{
				if(timespan > 0)
					log("FAILED: Transfered %.0f bytes in %.3f seconds: %.0f bps", bytesTransmitted, timespan, (bytesTransmitted*8/timespan));
				else
					log("FAILED: BT downstream transfer");
			}						
		}
		else if((exp == TCP) && (direction == DOWNSTREAM)){
			log("Running TCP downstream transfer port %d .", port);
			setNonblocking(fd);
			bytesTransmitted = connectionHandler(fd, duration, false, false, &start, &end);
			double timespan;
			if((end.tv_usec - start.tv_usec) >= 0)
				timespan = (end.tv_sec - start.tv_sec) + ((end.tv_usec - start.tv_usec)/1000000.0);
			else
				timespan = (end.tv_sec - start.tv_sec -1) + ((1000000 + end.tv_usec - start.tv_usec)/1000000.0);
			
			if(transferSuccessful)
				log("Transfered %.0f bytes in %.3f seconds: %.0f bps", bytesTransmitted, timespan, (bytesTransmitted*8/timespan));
			else{
				if(timespan > 0)
					log("FAILED: Transfered %.0f bytes in %.3f seconds: %.0f bps", bytesTransmitted, timespan, (bytesTransmitted*8/timespan));
				else
					log("FAILED: TCP downstream transfer");
			}
		}
		else if((exp == SETUP) && (direction == UPSTREAM)){
			log("Setting up upstream transfer.");
			tcpSender(fd, duration);
		}
		else if((exp == BT) && (direction == UPSTREAM)){
			log("Running BT upstream transfer port %d .", port);
			setNonblocking(fd);
			bytesTransmitted = connectionHandler(fd, duration, true, true, &start, &end);
			double timespan;
			if((end.tv_usec - start.tv_usec) >= 0)
				timespan = (end.tv_sec - start.tv_sec) + ((end.tv_usec - start.tv_usec)/1000000.0);
			else
				timespan = (end.tv_sec - start.tv_sec -1) + ((1000000 + end.tv_usec - start.tv_usec)/1000000.0);

			if(transferSuccessful)
				log("Transfered %.0f bytes in %.3f seconds: %.0f bps", bytesTransmitted, timespan, (bytesTransmitted*8/timespan));
			else{
				if(timespan > 0)
					log("FAILED: Transfered %.0f bytes in %.3f seconds: %.0f bps", bytesTransmitted, timespan, (bytesTransmitted*8/timespan));
				else
					log("FAILED: BT upstream transfer");
			}
		}
		else if((exp == TCP) && (direction == UPSTREAM)){
			log("Running TCP upstream transfer port %d .", port);
			setNonblocking(fd);
			bytesTransmitted = connectionHandler(fd, duration, false, true, &start, &end);

			double timespan;
			if((end.tv_usec - start.tv_usec) >= 0)
				timespan = (end.tv_sec - start.tv_sec) + ((end.tv_usec - start.tv_usec)/1000000.0);
			else
				timespan = (end.tv_sec - start.tv_sec -1) + ((1000000 + end.tv_usec - start.tv_usec)/1000000.0);

			if(transferSuccessful)
				log("Transfered %.0f bytes in %.3f seconds: %.0f bps", bytesTransmitted, timespan, (bytesTransmitted*8/timespan));
			else{
				if(timespan > 0)
					log("FAILED: Transfered %.0f bytes in %.3f seconds: %.0f bps", bytesTransmitted, timespan, (bytesTransmitted*8/timespan));
				else
					log("FAILED: TCP upstream transfer");
			}
		}
		
                log2x("Closing transfer socket");
                // Clean up
                close(fd);
                log2x("Closing listen socket");
                close(sock);
                log2x("Both sockets closed successfully"); 
	
		// Flush out captured packets
		pcap_dump_flush(myDumper);
	}
	
	// Listen for 2 seconds for more incoming packets
	gettimeofday(&tv, NULL);
	tv.tv_sec += 2;
	sniffPackets(-1, read_socket, &tv);

	if (myDumper){
		pcap_dump_close(myDumper);

		if(zip){
			// Compress dumped trace
			FILE *infile = fopen(dumpFile, "rb");
			if(infile){
				char gzdumpFile[503];	
				snprintf(gzdumpFile, 503, "%s.gz", dumpFile);
				gzFile outfile = gzopen(gzdumpFile, "wb");

				if(outfile){

					log("Zipping %s to %s", dumpFile, gzdumpFile);

					char inbuffer[128];
					int num_read = 0;
					while ((num_read = fread(inbuffer, 1, sizeof(inbuffer), infile)) > 0) {
						gzwrite(outfile, inbuffer, num_read);
					}
					fclose(infile);
					gzclose(outfile);		

					if(remove(dumpFile) != 0)
						perror( "Error deleting file" );
				}
			}
		}
	}
	

	// Send back an "ok\n" to the server that we are done.
	int ret = write(cs, "ok\n", 3);
	if(ret < 3)
		log("Error: Cannot write to command socket.");
	
	close(cs);
	close(command_port);
	
	log2("Done");
	fprintf(stderr, "bt_client: finished.\n");
}
